import warnings
from copy import deepcopy
from pathlib import Path
from typing import Any, Iterable, List

import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj import CRS
from shapely import geometry as gmt

from geost import config, spatial, utils
from geost._warnings import AlignmentWarning
from geost.abstract_classes import AbstractCollection, AbstractData, AbstractHeader
from geost.analysis import cumulative_thickness
from geost.export import (
    borehole_to_multiblock,
    export_to_dftgeodata,
    layerdata_to_pyvista_unstructured,
)
from geost.mixins import GeopandasExportMixin, PandasExportMixin
from geost.projections import (
    horizontal_reference_transformer,
    vertical_reference_transformer,
)
from geost.validation import safe_validate, schemas

type DataObject = DiscreteData | LayeredData
type HeaderObject = LineHeader | PointHeader
type Coordinate = int | float
type GeometryType = gmt.base.BaseGeometry | list[gmt.base.BaseGeometry]


class PointHeader(AbstractHeader, GeopandasExportMixin):
    def __init__(self, gdf, vertical_reference: str | int | CRS):
        self.gdf = gdf
        self.__vertical_reference = CRS(vertical_reference)

    def __repr__(self):
        name = self.__class__.__name__
        data = self._gdf
        length = len(data)
        return f"{name} instance containing {length} objects\n{data}"

    def __getitem__(self, column):
        return self.gdf[column]

    def __setitem__(self, key, values):
        self.gdf.loc[:, key] = values

    def __len__(self):
        return len(self.gdf)

    @property
    def gdf(self):
        """
        Underlying geopandas.GeodataFrame with header data.
        """
        return self._gdf

    @property
    def horizontal_reference(self):
        """
        Coordinate reference system represented by an instance of pyproj.crs.CRS
        """
        return self.gdf.crs

    @property
    def vertical_reference(self):
        """
        Vertical datum represented by an instance of pyproj.crs.CRS
        """
        return self.__vertical_reference

    @gdf.setter
    def gdf(self, gdf):
        self._gdf = safe_validate(schemas.pointheader, gdf)

    def change_horizontal_reference(self, to_epsg: str | int | CRS):
        """
        Change the horizontal reference (i.e. coordinate reference system, crs) of the
        header to the given target crs.

        Parameters
        ----------
        to_epsg : str | int | CRS
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input().

        Examples
        --------
        To change the header's current horizontal reference to WGS 84 UTM zone 31N:

        >>> self.change_horizontal_reference(32631)

        This would be the same as:

        >>> self.change_horizontal_reference("epsg:32631")

        As Pyproj is very flexible, you can even use the CRS's full official name:

        >>> self.change_horizontal_reference("WGS 84 / UTM zone 31N")

        """
        self._gdf = self.gdf.to_crs(to_epsg)
        self._gdf[["x", "y"]] = self._gdf[["x", "y"]].astype(float)
        self._gdf["x"] = self._gdf["geometry"].x
        self._gdf["y"] = self._gdf["geometry"].y

    def change_vertical_reference(self, to_epsg: str | int | CRS):
        """
        Change the vertical reference of the object's surface levels.

        Parameters
        ----------
        to_epsg : str | int | CRS
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum.

            Some often-used vertical datums are:
            - NAP : 5709
            - MSL NL depth : 9288
            - LAT NL depth : 9287
            - Ostend height : 5710

            See epsg.io for more.

        Examples
        --------
        To change the header's current vertical reference to NAP:

        >>> self.change_vertical_reference(5709)

        This would be the same as:

        >>> self.change_vertical_reference("epsg:5709")

        As the Pyproj constructors are very flexible, you can even use the CRS's full
        official name instead of an EPSG number. E.g. for changing to NAP and the
        Belgian Ostend height vertical datums repsectively, you can use:

        >>> self.change_vertical_reference("NAP")
        >>> self.change_vertical_reference("Ostend height")
        """
        transformer = vertical_reference_transformer(
            self.horizontal_reference, self.vertical_reference, to_epsg
        )
        self.gdf[["surface", "end"]] = self.gdf[["surface", "end"]].astype(float)
        _, _, new_surface = transformer.transform(
            self.gdf["x"], self.gdf["y"], self.gdf["surface"]
        )
        _, _, new_end = transformer.transform(
            self.gdf["x"], self.gdf["y"], self.gdf["end"]
        )
        self._gdf.loc[:, "surface"] = new_surface
        self._gdf.loc[:, "end"] = new_end
        self.__vertical_reference = CRS(to_epsg)

    def get(self, selection_values: str | Iterable, column: str = "nr"):
        """
        Get a subset of a header through a string or iterable of object id(s).
        Optionally uses a different column than "nr" (the column with object ids).

        Parameters
        ----------
        selection_values : str | Iterable
            Values to select.
        column : str, optional
            In which column of the header to look for selection values, by default "nr".

        Returns
        -------
        Instance of :class:`~geost.base.PointHeader`.
            Instance of :class:`~geost.base.PointHeader` containing only
            objects selected through this method.

        Examples
        --------
        >>> self.get(["obj1", "obj2"])

        will return a collection with only these objects.

        Suppose we have a number of boreholes that we have joined with geological
        map units using the method
        :meth:`~geost.base.PointHeader.get_area_labels`. We have added this data
        to the header table in the column 'geological_unit'. Using:

        >>> self.get(["unit1", "unit2"], column="geological_unit")

        will return a :class:`~geost.base.PointHeader` with all boreholes
        that are located in "unit1" and "unit2" geological map areas.
        """
        if isinstance(selection_values, str):
            selected_gdf = self[self[column] == selection_values]
        elif isinstance(selection_values, Iterable):
            selected_gdf = self[self[column].isin(selection_values)]

        selected_gdf = selected_gdf[~selected_gdf.duplicated()]

        return self.__class__(selected_gdf, self.vertical_reference)

    def select_within_bbox(
        self,
        xmin: Coordinate,
        ymin: Coordinate,
        xmax: Coordinate,
        ymax: Coordinate,
        invert: bool = False,
    ):
        """
        Make a selection of the header based on a bounding box.

        Parameters
        ----------
        xmin : float | int
            Minimum x-coordinate of the bounding box.
        ymin : float | int
            Minimum y-coordinate of the bounding box.
        xmax : float | int
            Maximum x-coordinate of the bounding box.
        ymax : float | int
            Maximum y-coordinate of the bounding box.
        invert : bool, optional
            Invert the selection, so select all objects outside of the
            bounding box in this case, by default False.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            Instance of :class:`~geost.base.PointHeader` containing only selected
            geometries.
        """
        gdf_selected = spatial.select_points_within_bbox(
            self.gdf, xmin, ymin, xmax, ymax, invert=invert
        )
        return self.__class__(gdf_selected, self.vertical_reference)

    def select_with_points(
        self,
        points: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int,
        invert: bool = False,
    ):
        """
        Make a selection of the header based on point geometries.

        Parameters
        ----------
        points : str | Path | gpd.GeoDataFrame | GeometryType
            Any type of point geometries that can be used for the selection: GeoDataFrame
            containing points or filepath to a shapefile like file, or Shapely Point,
            MultiPoint or list containing Point objects.
        buffer : float | int
            Buffer distance for selection geometries.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            Instance of :class:`~geost.base.PointHeader` containing only selected
            geometries.

        """
        gdf_selected = spatial.select_points_near_points(
            self.gdf, points, buffer, invert=invert
        )
        return self.__class__(gdf_selected, self.vertical_reference)

    def select_with_lines(
        self,
        lines: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int,
        invert: bool = False,
    ):
        """
        Make a selection of the header based on line geometries.

        Parameters
        ----------
        lines : str | Path | gpd.GeoDataFrame | GeometryType
            Any type of line geometries that can be used for the selection: GeoDataFrame
            containing lines or filepath to a shapefile like file, or Shapely LineString,
            MultiLineString or list containing LineString objects.
        buffer : float | int
            Buffer distance for selection geometries.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            Instance of :class:`~geost.base.PointHeader` containing only selected
            geometries.

        """
        gdf_selected = spatial.select_points_near_lines(
            self.gdf, lines, buffer, invert=invert
        )
        return self.__class__(gdf_selected, self.vertical_reference)

    def select_within_polygons(
        self,
        polygons: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int = 0,
        invert: bool = False,
    ):
        """
        Make a selection of the header based on polygon geometries.

        Parameters
        ----------
        polygons : str | Path | gpd.GeoDataFrame | GeometryType
            Any type of polygon geometries that can be used for the selection: GeoDataFrame
            containing polygons or filepath to a shapefile like file, or Shapely Polygon,
            MultiPolygon or list containing Polygon objects.
        buffer : float | int, optional
            Optional buffer distance around the polygon selection geometries, by default
            0.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            Instance of :class:`~geost.base.PointHeader` containing only selected
            geometries.

        """
        gdf_selected = spatial.select_points_within_polygons(
            self.gdf, polygons, buffer, invert=invert
        )
        return self.__class__(gdf_selected, self.vertical_reference)

    def select_by_depth(
        self,
        top_min: float = None,
        top_max: float = None,
        end_min: float = None,
        end_max: float = None,
    ):
        """
        Select data from depth constraints. If a keyword argument is not given it will
        not be considered. e.g. if you need only boreholes that go deeper than -500 m
        use only end_max = -500.

        Parameters
        ----------
        top_min : float, optional
            Minimum elevation of the borehole/cpt top, by default None.
        top_max : float, optional
            Maximum elevation of the borehole/cpt top, by default None.
        end_min : float, optional
            Minimum elevation of the borehole/cpt end, by default None.
        end_max : float, optional
            Maximumelevation of the borehole/cpt end, by default None.

        Returns
        -------
        Child of :class:`~geost.base.PointHeader`.
            Instance of :class:`~geost.base.PointHeader` or containing only objects
            selected by this method.
        """
        selected = self.gdf.copy()
        if top_min is not None:
            selected = selected[selected["surface"] >= top_min]
        if top_max is not None:
            selected = selected[selected["surface"] <= top_max]
        if end_min is not None:
            selected = selected[selected["end"] >= end_min]
        if end_max is not None:
            selected = selected[selected["end"] <= end_max]

        selected = selected[~selected.duplicated()]

        return self.__class__(selected, self.vertical_reference)

    def select_by_length(self, min_length: float = None, max_length: float = None):
        """
        Select data from length constraints: e.g. all boreholes between 50 and 150 m
        long. If a keyword argument is not given it will not be considered.

        Parameters
        ----------
        min_length : float, optional
            Minimum length of borehole/cpt, by default None.
        max_length : float, optional
            Maximum length of borehole/cpt, by default None.

        Returns
        -------
        Child of :class:`~geost.base.PointHeader`.
            Instance of :class:`~geost.base.PointHeader` or containing only objects
            selected by this method.
        """
        selected = self.gdf.copy()
        length = selected["surface"] - selected["end"]
        if min_length is not None:
            selected = selected[length >= min_length]
        if max_length is not None:
            selected = selected[length <= max_length]

        selected = selected[~selected.duplicated()]

        return self.__class__(selected, self.vertical_reference)

    def get_area_labels(
        self,
        polygon_gdf: str | Path | gpd.GeoDataFrame,
        column_name: str | Iterable,
        include_in_header=False,
    ) -> pd.DataFrame:
        """
        Find in which area (polygons) the point data locations fall. e.g. to determine
        in which geomorphological unit points are located.

        Parameters
        ----------
        polygon_gdf : str | Path | gpd.GeoDataFrame
            GeoDataFrame with polygons.
        column_name : str | Iterable
            The column name to find the labels in. Given as a string or iterable of
            strings in case you'd like to find multiple labels.
        include_in_header : bool, optional
            Whether to add the acquired data to the header table or not, by default
            False.

        Returns
        -------
        pd.DataFrame
            Borehole ids and the polygon label they are in. If include_in_header = True,
            a column containing the generated data will be added inplace.
        """
        polygon_gdf = utils.check_geometry_instance(polygon_gdf)
        polygon_gdf = spatial.check_and_coerce_crs(
            polygon_gdf, self.horizontal_reference
        )

        all_nrs = self["nr"]
        area_labels = spatial.find_area_labels(self.gdf, polygon_gdf, column_name)
        area_labels = pd.concat([all_nrs, area_labels], axis=1)

        if include_in_header:
            self.gdf.drop(
                columns=column_name,
                errors="ignore",
                inplace=True,
            )
            self._gdf = self.gdf.merge(area_labels, on="nr")
        else:
            return area_labels


class LineHeader(AbstractHeader, GeopandasExportMixin):  # pragma: no cover
    def __init__(self, gdf, vertical_reference: str | int | CRS):
        self.gdf = gdf
        self.__vertical_reference = CRS(vertical_reference)

    def __repr__(self):
        return f"{self.__class__.__name__} instance containing {len(self)} objects"

    def __getitem__(self, column):
        return self.gdf[column]

    def __setitem__(self, key, values):
        self.gdf.loc[:, key] = values

    def __len__(self):
        return len(self.gdf)

    @property
    def gdf(self):
        return self._gdf

    @property
    def horizontal_reference(self):
        return self.gdf.crs

    @property
    def vertical_reference(self):
        return self.__vertical_reference

    @gdf.setter
    def gdf(self, gdf):
        self._gdf = safe_validate(schemas.lineheader, gdf)

    def change_horizontal_reference(self, to_epsg: str | int | CRS):
        raise NotImplementedError("Add function logic")

    def change_vertical_reference(self, to_epsg: str | int | CRS):
        raise NotImplementedError("Add function logic")

    def get(self):
        raise NotImplementedError("Add function logic")

    def select_within_bbox(self):
        raise NotImplementedError("Add function logic")

    def select_with_points(self):
        raise NotImplementedError("Add function logic")

    def select_with_lines(self):
        raise NotImplementedError("Add function logic")

    def select_within_polygons(self):
        raise NotImplementedError("Add function logic")

    def select_by_depth(self):
        raise NotImplementedError("Add function logic")

    def select_by_length(self):
        raise NotImplementedError("Add function logic")

    def get_area_labels(self):
        raise NotImplementedError("Add function logic")


class LayeredData(AbstractData, PandasExportMixin):
    """
    A class to hold layered data objects (i.e. containing "tops" and "bottoms") like
    borehole descriptions which can be used for selections and exports.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame containing the data. Mandatory columns that must be present in the
        DataFrame are: "nr", "x", "y", "surface", "top" and "bottom". Otherwise, many methods
        in the class will not work.
    has_inclined : bool, optional
        If True, the data also contains inclined objects which means the top of layers is
        not in the same x,y-location as the bottom of layers.

    """

    __datatype = "layered"

    def __init__(
        self,
        df: pd.DataFrame,
        has_inclined: bool = False,
    ):
        self.has_inclined = has_inclined
        self.df = df

    def __repr__(self):
        name = self.__class__.__name__
        data = self._df
        return f"{name} instance:\n{data}"

    def __getitem__(self, column):
        return self.df[column]

    def __setitem__(self, column, item):
        self.df.loc[:, column] = item

    def __len__(self):
        return len(self.df)

    @property
    def df(self):
        return self._df

    @property
    def datatype(self):
        return self.__datatype

    @df.setter
    def df(self, df):
        """
        Underlying pandas.DataFrame, validated upon setting the attr.
        """
        self._df = (
            safe_validate(schemas.layerdata_inclined, df)
            if self.has_inclined
            else safe_validate(schemas.layerdata, df)
        )

    @staticmethod
    def _check_correct_instance(selection_values: str | Iterable) -> Iterable:
        if isinstance(selection_values, str):
            selection_values = [selection_values]
        return selection_values

    @staticmethod
    def _change_depth_values(df: pd.DataFrame) -> pd.DataFrame:
        df.loc[:, "top"] = df["surface"] - df["top"]
        df.loc[:, "bottom"] = df["surface"] - df["bottom"]
        return df

    def to_header(
        self,
        horizontal_reference: str | int | CRS = 28992,
        vertical_reference: str | int | CRS = 5709,
    ):
        """
        Generate a :class:`~geost.base.PointHeader` from this instance of LayaredData.

        Parameters
        ----------
        horizontal_reference : str | int | CRS, optional
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(), by default 28992.
        vertical_reference : str | int | CRS, optional
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
            "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
            5710, by default 5709.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            An instance of :class:`~geost.base.PointHeader`
        """
        header_columns = ["nr", "x", "y", "surface", "end"]
        header = self[header_columns].drop_duplicates("nr").reset_index(drop=True)
        header = utils.dataframe_to_geodataframe(header).set_crs(horizontal_reference)
        return PointHeader(header, vertical_reference)

    def to_collection(
        self,
        horizontal_reference: str | int | CRS = 28992,
        vertical_reference: str | int | CRS = 5709,
    ):
        """
        Create a collection from this instance of LayeredData. A collection combines
        header and data and ensures that they remain aligned when applying methods.

        Parameters
        ----------
        horizontal_reference : str | int | CRS, optional
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(), by default 28992.
        vertical_reference : str | int | CRS, optional
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
            "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
            5710, by default 5709.

        Returns
        -------
        :class:`~geost.base.Collection`
            An instance of :class:`~geost.base.Collection`
        """
        header = self.to_header(horizontal_reference, vertical_reference)
        return BoreholeCollection(header, self)
        # NOTE: Type of Collection may need to be inferred in the future.

    def select_by_values(
        self, column: str, selection_values: str | Iterable, how: str = "or"
    ):
        """
        Select data based on the presence of given values in a given column. Can be used
        for example to select boreholes that contain peat in the lithology column.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data to use when looking for
            values.
        selection_values : str | Iterable
            Value or values to look for in the column.
        how : str, optional
            Either "and" or "or". "and" requires all selection values to be present in
            column for selection. "or" will select the core if any one of the
            selection_values are found in the column. Default is "and".

        Returns
        -------
        :class:`~geost.base.LayeredData`
            New instance containing only the data selected by this method.

        Examples
        --------
        To select boreholes where both clay ("K") and peat ("V") are present at the same
        time, use "and" as a selection method:

        >>> boreholes.select_by_values("lith", ["V", "K"], how="and")

        To select boreholes that can have one, or both lithologies, use or as the selection
        method:

        >>> boreholes.select_by_values("lith", ["V", "K"], how="and")

        """
        if column not in self.df.columns:
            raise IndexError(
                f"The column '{column}' does not exist and cannot be used for selection"
            )

        if isinstance(selection_values, str):
            selection_values = [selection_values]

        selected = self.df
        if how == "or":
            valid = self["nr"][self[column].isin(selection_values)].unique()
            selected = selected[selected["nr"].isin(valid)]

        elif how == "and":
            for value in selection_values:
                valid = self["nr"][self[column] == value].unique()
                selected = selected[selected["nr"].isin(valid)]

        return self.__class__(selected, self.has_inclined)

    def slice_depth_interval(
        self,
        upper_boundary: float | int = None,
        lower_boundary: float | int = None,
        relative_to_vertical_reference: bool = False,
        update_layer_boundaries: bool = True,
    ):
        """
        Slice data based on given upper and lower boundaries. This returns a new object
        containing only the sliced data.

        Parameters
        ----------
        upper_boundary : float | int, optional
            Every layer that starts above this is removed. The default is None.
        lower_boundary : float | int, optional
            Every layer that starts below this is removed. The default is None.
        relative_to_vertical_reference : bool, optional
            If True, the slicing is done with respect to any kind of vertical reference
            plane (e.g. "NAP", "TAW"). If False, the slice is done with respect to depth
            below the surface. The default is False.
        update_layer_boundaries : bool, optional
            If True, the layer boundaries in the sliced data are updated according to the
            upper and lower boundaries used with the slice. If False, the original layer
            boundaries are kept in the sliced object. The default is False.

        Returns
        -------
        :class:`~geost.base.LayeredData`
            New instance containing only the data selected by this method.

        Examples
        --------
        Usage depends on whether the slicing is done with respect to depth below the
        surface or to a vertical reference plane.

        For example, select layers in data that are between 2 and 3 meters below the
        surface:

        >>> data.slice_depth_interval(2, 3)

        By default, the method updates the layer boundaries in sliced object according to
        the upper and lower boundaries. To suppress this behaviour use:

        >>> data.slice_depth_interval(2, 3, update_layer_boundaries=False)

        Slicing can also be done with respect to a vertical reference plane like "NAP".
        For example, to select layers in boreholes that are between -3 and -5 m NAP, use:

        >>> data.slice_depth_interval(-3, -5, relative_to_vertical_reference=True)

        """
        if not upper_boundary:
            upper_boundary = 1e34 if relative_to_vertical_reference else -1e34

        if not lower_boundary:
            lower_boundary = -1e34 if relative_to_vertical_reference else 1e34

        sliced = self.df.copy()

        if relative_to_vertical_reference:
            bounds_are_series = True
            upper_boundary = self["surface"] - upper_boundary
            lower_boundary = self["surface"] - lower_boundary
        else:
            bounds_are_series = False

        sliced = sliced[
            (sliced["bottom"] > upper_boundary) & (sliced["top"] < lower_boundary)
        ]

        if update_layer_boundaries:
            if bounds_are_series:
                upper_boundary = upper_boundary.loc[sliced.index]
                lower_boundary = lower_boundary.loc[sliced.index]

            sliced.loc[sliced["top"] <= upper_boundary, "top"] = upper_boundary
            sliced.loc[sliced["bottom"] >= lower_boundary, "bottom"] = lower_boundary

        return self.__class__(sliced, self.has_inclined)

    def slice_by_values(
        self, column: str, selection_values: str | Iterable, invert: bool = False
    ):
        """
        Slice rows from data based on matching condition. E.g. only return rows with
        a certain lithology in the collection object.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data to use when looking for
            values.
        selection_values : str | Iterable
            Values to look for in the column.
        invert : bool, optional
            If True, invert the slicing action, so remove layers with selected values
            instead of keeping them. The default is False.

        Returns
        -------
        :class:`~geost.base.LayeredData`
            New instance containing only the data objects selected by this method.

        Examples
        --------
        Return only rows in data contain sand ("Z") as lithology:

        >>> data.slice_by_values("lith", "Z")

        If you want all the rows that may contain everything but sand, use the "invert"
        option:

        >>> data.slice_by_values("lith", "Z", invert=True)

        """
        selection_values = self._check_correct_instance(selection_values)

        sliced = self.df.copy()

        if invert:
            sliced = sliced[~sliced[column].isin(selection_values)]
        else:
            sliced = sliced[sliced[column].isin(selection_values)]

        return self.__class__(sliced, self.has_inclined)

    def select_by_condition(self, condition: Any, invert: bool = False):
        """
        Select data using a manual condition that results in a boolean mask. Returns the
        rows in the data where the 'condition' evaluates to True.

        Parameters
        ----------
        condition : list, pd.Series or array like
            Boolean array like object with locations at which the values will be
            preserved, dtype must be 'bool' and the length must correspond with the
            length of the data.
        invert : bool, optional
            If True, the selection is inverted so rows that evaluate to False will be
            returned. The default is False.

        Returns
        -------
        :class:`~geost.base.LayeredData`
            New instance containing only the data objects selected by this method.

        Examples
        --------
        Select rows in data that contain a specific value:

        >>> data.select_by_condition(data["lith"]=="V")

        Or select rows in the data that contain a specific (part of) string or strings:

        >>> data.select_by_condition(data["column"].str.contains("foo|bar"))

        """
        if invert:
            selected = self[~condition]
        else:
            selected = self[condition]
        return self.__class__(selected, self.has_inclined)

    def get_cumulative_thickness(self, column: str, values: str | List[str]):
        """
        Get the cumulative thickness of layers where a column contains a specified search
        value or values.

        Parameters
        ----------
        column : str
            Name of column that must contain the search value or values.
        values : str | List[str]
            Search value or values in the column to find the cumulative thickness for.

        Returns
        -------
        pd.DataFrame
            Borehole ids and cumulative thickness of selected layers.

        Examples
        --------
        Get the cumulative thickness of the layers with lithology "K" in the column "lith"
        use:

        >>> data.get_cumulative_thickness("lith", "K")

        Or get the cumulative thickness for multiple selection values. In this case, a
        Pandas DataFrame is returned with a column per selection value containing the
        cumulative thicknesses:

        >>> data.get_cumulative_thickness("lith", ["K", "Z"])

        """
        selected_layers = self.slice_by_values(column, values)
        cum_thickness = selected_layers.df.groupby(["nr", column]).apply(
            cumulative_thickness, include_groups=False
        )
        cum_thickness = cum_thickness.unstack(level=column)
        return cum_thickness

    def get_layer_top(
        self,
        column: str,
        values: str | List[str],
        min_thickness: float = 0,
        min_depth: float = 0,
    ):
        """
        Find the depth at which a specified layer first occurs, starting at min_depth
        and looking downwards until the first layer of min_thickness is found of the
        specified layer.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data.
        values : str | List[str]
            Value or values of entries in the column that you want to find top of.
        min_thickness : float, optional
            Minimal thickness of the layer to be considered. The default is 0.
        min_depth : float, optional
            Minimal depth of the layer to be considered. The default is 0.

        Returns
        -------
        pd.DataFrame
            Borehole ids and top levels of selected layers in meters below the surface.

        Examples
        --------
        Get the top depth of layers in data where the lithology in the "lith" column
        is sand ("Z"):

        >>> data.get_layer_top("lith", "Z")

        """
        selected_layers = self.slice_by_values(column, values)
        selected_layers = selected_layers.slice_depth_interval(upper_boundary=min_depth)
        selected_layers = selected_layers.select_by_condition(
            selected_layers["bottom"] - selected_layers["top"] >= min_thickness
        )
        layer_top = selected_layers.df.groupby(["nr", column])["top"].first()
        return layer_top.unstack(level=column)

    def to_pyvista_cylinders(
        self,
        displayed_variables: str | List[str],
        radius: float = 1,
        vertical_factor: float = 1.0,
        relative_to_vertical_reference: bool = True,
    ):
        """
        Create a Pyvista MultiBlock object of cylinder-shaped geometries to represent
        boreholes. Although cylinders are prettier when visualized, they are quite costly
        to render in large numbers. Consider using
        :meth:`~geost.base.LayeredData.to_pyvista_grid` instead for large datasets.

        Parameters
        ----------
        displayed_variables : str | List[str]
            Name or names of data columns to include for visualisation. Can be columns that
            contain an array of floats, ints and strings.
        radius : float, optional
            Radius of the cylinders in m in the MultiBlock. The default is 1.
        vertical_factor : float, optional
            Factor to correct vertical scale. For example, when layer boundaries are given
            in cm, use 0.01 to convert to m. The default is 1.0, so no correction is applied.
            It is not recommended to use this for vertical exaggeration, use viewer functionality
            for that instead.
        relative_to_vertical_reference : bool, optional
            If True, the depth of the objects in the vtm file will be with respect to a
            reference plane (e.g. "NAP", "TAW"). If False, the depth will be with respect
            to 0.0. The default is True.

        Returns
        -------
        pyvista.MultiBlock
            A composite class holding the data which can be iterated over.

        """
        data_columns = self._check_correct_instance(displayed_variables)

        data = self.df.copy()

        if relative_to_vertical_reference:
            data = self._change_depth_values(data)
        else:
            data["surface"] = 0

        vtk_object = borehole_to_multiblock(data, data_columns, radius, vertical_factor)
        return vtk_object

    def to_pyvista_grid(
        self,
        displayed_variables: str | list[str],
        radius: float = 1,
    ):
        """
        Create a PyVista UnstructuredGrid object of the data in this instance. This
        method is more efficient than :meth:`~geost.base.LayeredData.to_pyvista_cylinders`
        for large datasets, as it uses a grid representation instead of cylinders.

        Parameters
        ----------
        data_columns : str | list[str]
            Name or names of data columns to include for visualisation. Can be columns that
            contain an array of floats, ints and strings.
        radius : float, optional
            Radius cells in m, by default 1.

        Returns
        -------
        pyvista.UnstructuredGrid
            A PyVista UnstructuredGrid object containing the data that can be used for
            3D visualisation in PyVista or other VTK viewers.

        """
        displayed_variables = self._check_correct_instance(displayed_variables)
        vtk_object = layerdata_to_pyvista_unstructured(
            self.df, displayed_variables, radius=radius
        )
        return vtk_object

    def to_datafusiontools(
        self,
        columns: List[str],
        outfile: str | Path = None,
        encode: bool = False,
        relative_to_vertical_reference: bool = True,
    ):
        """
        Export all data to the core "Data" class of Deltares DataFusionTools. Returns
        a list of "Data" objects, one for each data object that is exported. This list
        can directly be used within DataFusionTools. If out_file is given, the list of
        Data objects is saved to a pickle file.

        For DataFusionTools visit:
        https://bitbucket.org/DeltaresGEO/datafusiontools/src/master/

        Parameters
        ----------
        columns : List[str]
            Which columns in the data to include for the export. These will become variables
            in the DataFusionTools "Data" class.
        outfile : str | Path, optional
            If a path to outfile is given, the data is written to a pickle file.
        encode : bool, default True
            If True, categorical data columns are encoded to additional binary columns
            (all possible values become a seperate feature that is 0 or 1). The default is
            False. Warning: if there is a large number of possible categories, many columns
            with categorical data or both, the export process may become slow and may consume
            a large amount memory. Please consider carefully which categorical data columns
            need to be included.
        relative_to_vertical_reference : bool, optional
            If True, the depth of all data objects will converted to a depth with respect to
            a reference plane (e.g. "NAP", "TAW"). If False, the depth will be kept as original
            in the "top" and "bottom" columns which is in meter below the surface. The default
            is True.

        Returns
        -------
        List[Data]
            List containing the DataFusionTools Data objects.

        """
        columns = self._check_correct_instance(columns)
        data = self.df.copy()
        if relative_to_vertical_reference:
            data = self._change_depth_values(data)

        dftgeodata = export_to_dftgeodata(data, columns, encode=encode)

        if outfile:
            utils.save_pickle(dftgeodata, outfile)
        else:
            return dftgeodata

    def _create_geodataframe_3d(
        self, relative_to_vertical_reference: bool = True, crs: str | int | CRS = None
    ):
        """
        Helper method for export method "to_qgis3d" to create the necessary GeoDataFrame
        containing 3D Shapely objects and associated information.

        Parameters
        ----------
        relative_to_vertical_reference : bool, optional
            If True, the depth of all data objects will converted to a depth with respect to
            a reference plane (e.g. "NAP", "TAW"). If False, the depth will be kept as original
            in the "top" and "bottom" columns which is in meter below the surface. The default
            is True.
        crs : str | int | CRS
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input().
        """
        data = self.df.copy()

        if relative_to_vertical_reference:
            data = self._change_depth_values(data)

        data_columns = [
            col
            for col in data.columns
            if col
            not in ["nr", "x", "y", "x_bot", "y_bot", "surface", "end", "top", "bottom"]
        ]

        data_to_write = dict(
            nr=data["nr"].values,
            top=data["top"].values.astype(float),
            bottom=data["bottom"].values.astype(float),
        )

        data_to_write.update(data[data_columns].to_dict(orient="list"))

        if self.has_inclined:
            geometries = [
                gmt.LineString([[x, y, top + 0.01], [x_bot, y_bot, bottom + 0.01]])
                for x, y, x_bot, y_bot, top, bottom in zip(
                    data["x"].values.astype(float),
                    data["y"].values.astype(float),
                    data["x_bot"].values.astype(float),
                    data["y_bot"].values.astype(float),
                    data["top"].values.astype(float),
                    data["bottom"].values.astype(float),
                )
            ]
        else:  # NOTE: Doesn't it need to be "top - 0.01" to create overlap?
            geometries = [
                gmt.LineString([[x, y, top + 0.01], [x, y, bottom + 0.01]])
                for x, y, top, bottom in zip(
                    data["x"].values.astype(float),
                    data["y"].values.astype(float),
                    data["top"].values.astype(float),
                    data["bottom"].values.astype(float),
                )
            ]

        gdf = gpd.GeoDataFrame(
            data=data_to_write,
            geometry=geometries,
            crs=crs,
        )
        return gdf

    def to_qgis3d(
        self,
        outfile: str | Path,
        relative_to_vertical_reference: bool = True,
        crs: str | int | CRS = None,
        **kwargs,
    ):
        """
        Write data to geopackage file that can be directly loaded in the Qgis2threejs
        plugin. Works only for layered (borehole) data.

        Parameters
        ----------
        outfile : str | Path
            Path to geopackage file to be written.
        relative_to_vertical_reference : bool, optional
            If True, the depth of all data objects will converted to a depth with
            respect to a reference plane (e.g. "NAP", "Ostend height"). If False, the
            depth will be kept as original in the "top" and "bottom" columns which is in
            meter below the surface. The default is True.
        crs : str | int | CRS
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input().

        **kwargs
            geopandas.GeodataFrame.to_file kwargs. See relevant Geopandas documentation.

        """
        qgis3d = self._create_geodataframe_3d(relative_to_vertical_reference, crs=crs)
        qgis3d.to_file(outfile, driver="GPKG", **kwargs)

    def to_kingdom(
        self,
        outfile: str | Path,
        tdstart: int = 1,
        vw: float = 1500.0,
        vs: float = 1600.0,
    ):
        """
        Write data to 2 csv files: 1) interval data and 2) time-depth chart. These files
        can be imported in the Kingdom seismic interpretation software.

        Parameters
        ----------
        outfile : str | Path
            Path to csv file to be written.
        tdstart : int
            startindex for TDchart, default is 1
        vw : float
            sound velocity in water in m/s, default is 1500 m/s
        vs : float
            sound velocity in sediment in m/s, default is 1600 m/s
        """
        # 1. add column needed in kingdom and write interval data
        kingdom_df = self.df.copy()
        # Add total depth and rename bottom and top columns to Kingdom requirements
        kingdom_df.insert(7, "Total depth", (kingdom_df["surface"] - kingdom_df["end"]))
        kingdom_df.rename(
            columns={"top": "Start depth", "bottom": "End depth"}, inplace=True
        )
        kingdom_df.to_csv(outfile, index=False)

        # 2. create and write time-depth chart
        tdchart = self[["nr", "surface"]].copy()
        tdchart.drop_duplicates(inplace=True)
        tdchart.insert(0, "id", range(tdstart, tdstart + len(tdchart)))
        # Add measured depth (predefined depths of 0 and 1 m below surface)
        tdchart = pd.concat(
            [
                tdchart.assign(MD=np.zeros(len(tdchart), dtype=np.int64)),
                tdchart.assign(MD=np.ones(len(tdchart), dtype=np.int64)),
            ]
        )
        # Add two-way travel time
        tdchart["TWT"] = (-tdchart["surface"] / (vw / 2 / 1000)) + (
            tdchart["MD"] * 1 / (vs / 2 / 1000)
        )

        tdchart.drop("surface", axis=1, inplace=True)
        tdchart.sort_values(by=["id", "MD"], inplace=True)
        if not isinstance(outfile, Path):
            outfile = Path(outfile)
        tdchart.to_csv(
            outfile.parent.joinpath(f"{outfile.stem}_TDCHART{outfile.suffix}"),
            index=False,
        )


class DiscreteData(AbstractData, PandasExportMixin):
    __datatype = "discrete"

    def __init__(self, df, has_inclined: bool = False):
        self.has_inclined = has_inclined
        self.df = df

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, df):
        self._df = (
            safe_validate(schemas.discretedata_inclined, df)
            if self.has_inclined
            else safe_validate(schemas.discretedata, df)
        )

    def __repr__(self):
        name = self.__class__.__name__
        data = self._df
        return f"{name} instance:\n{data}"

    def __getitem__(self, column):
        return self.df[column]

    def __setitem__(self, column, item):
        self.df.loc[:, column] = item

    def __len__(self):
        return len(self.df)

    @property
    def datatype(self):
        return self.__datatype

    def to_header(
        self,
        horizontal_reference: str | int | CRS = 28992,
        vertical_reference: str | int | CRS = 5709,
    ):
        """
        Generate a :class:`~geost.base.PointHeader` from this instance of DiscreteData.

        Parameters
        ----------
        horizontal_reference : str | int | CRS, optional
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(), by default 28992.
        vertical_reference : str | int | CRS, optional
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
            "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
            5710, by default 5709.

        Returns
        -------
        :class:`~geost.base.PointHeader`
            An instance of :class:`~geost.base.PointHeader`

        """
        header_columns = ["nr", "x", "y", "surface", "end"]
        header = self[header_columns].drop_duplicates("nr").reset_index(drop=True)
        header = utils.dataframe_to_geodataframe(header).set_crs(horizontal_reference)
        return PointHeader(header, vertical_reference)

    def to_collection(
        self,
        horizontal_reference: str | int | CRS = 28992,
        vertical_reference: str | int | CRS = 5709,
    ):
        """
        Create a collection from this instance of DiscreteData. A collection combines
        header and data and ensures that they remain aligned when applying methods.

        Parameters
        ----------
        horizontal_reference : str | int | CRS, optional
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(), by default 28992.
        vertical_reference : str | int | CRS, optional
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
            "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
            5710, by default 5709.

        Returns
        -------
        :class:`~geost.base.Collection`
            An instance of :class:`~geost.base.Collection`
        """
        header = self.to_header(horizontal_reference, vertical_reference)
        return CptCollection(
            header, self
        )  # TODO: type of Collection needs to be inferred in the future

    def select_by_values(
        self, column: str, selection_values: str | Iterable, how: str = "or"
    ):
        """
        Select data based on the presence of given values in a given column containing
        categorical data. Can be used for example to select points that contain peat in
        the lithology column.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data to use when looking for
            values.
        selection_values : str | Iterable
            Value or values to look for in the column.
        how : str, optional
            Either "and" or "or". "and" requires all selection values to be present in
            column for selection. "or" will select the core if any one of the
            selection_values are found in the column. Default is "and".

        Returns
        -------
        :class:`~geost.base.DiscreteData`
            New instance containing only the data selected by this method.

        Examples
        --------
        To select data where both clay ("K") and peat ("V") are present at the same time,
        use "and" as a selection method:

        >>> data.select_by_values("lith", ["V", "K"], how="and")

        To select data that can have one, or both lithologies, use or as the selection
        method:

        >>> data.select_by_values("lith", ["V", "K"], how="and")

        """
        if column not in self.df.columns:
            raise IndexError(
                f"The column '{column}' does not exist and cannot be used for selection"
            )

        if isinstance(selection_values, str):
            selection_values = [selection_values]

        selected = self.df.copy()
        if how == "or":
            valid = self["nr"][self[column].isin(selection_values)].unique()
            selected = selected[selected["nr"].isin(valid)]

        elif how == "and":
            for value in selection_values:
                valid = self["nr"][self[column] == value].unique()
                selected = selected[selected["nr"].isin(valid)]

        return self.__class__(selected, self.has_inclined)

    def slice_depth_interval(
        self,
        upper_boundary: float | int = None,
        lower_boundary: float | int = None,
        relative_to_vertical_reference: bool = False,
    ):
        """
        Slice data based on given upper and lower boundaries. This returns a new object
        containing only the sliced data.

        Parameters
        ----------
        upper_boundary : float | int, optional
            Every layer that starts above this is removed. The default is None.
        lower_boundary : float | int, optional
            Every layer that starts below this is removed. The default is None.
        relative_to_vertical_reference : bool, optional
            If True, the slicing is done with respect to any kind of vertical reference
            plane (e.g. "NAP", "TAW"). If False, the slice is done with respect to depth
            below the surface. The default is False.
        update_layer_boundaries : bool, optional
            If True, the layer boundaries in the sliced data are updated according to the
            upper and lower boundaries used with the slice. If False, the original layer
            boundaries are kept in the sliced object. The default is False.

        Returns
        -------
        :class:`~geost.base.DiscreteData`
            New instance containing only the data selected by this method.

        Examples
        --------
        Usage depends on whether the slicing is done with respect to depth below the
        surface or to a vertical reference plane.

        For example, select layers in data that are between 2 and 3 meters below the
        surface:

        >>> data.slice_depth_interval(2, 3)

        Slicing can also be done with respect to a vertical reference plane like "NAP".
        For example, to select layers in boreholes that are between -3 and -5 m NAP, use:

        >>> data.slice_depth_interval(-3, -5, relative_to_vertical_reference=True)

        """
        if not upper_boundary:
            upper_boundary = 1e34 if relative_to_vertical_reference else -1e34

        if not lower_boundary:
            lower_boundary = -1e34 if relative_to_vertical_reference else 1e34

        sliced = self.df.copy()

        if relative_to_vertical_reference:
            upper_boundary = self["surface"] - upper_boundary
            lower_boundary = self["surface"] - lower_boundary

        sliced = sliced[
            (sliced["depth"] >= upper_boundary) & (sliced["depth"] <= lower_boundary)
        ]

        return self.__class__(sliced, self.has_inclined)

    def slice_by_values(self):  # pragma: no cover
        raise NotImplementedError()

    def select_by_condition(self, condition: Any, invert: bool = False):
        """
        Select data using a manual condition that results in a boolean mask. Returns the
        rows in the data where the 'condition' evaluates to True.

        Parameters
        ----------
        condition : list, pd.Series or array like
            Boolean array like object with locations at which the values will be
            preserved, dtype must be 'bool' and the length must correspond with the
            length of the data.
        invert : bool, optional
            If True, the selection is inverted so rows that evaluate to False will be
            returned. The default is False.

        Returns
        -------
        :class:`~geost.base.DiscreteData`
            New instance containing only the data objects selected by this method.

        Examples
        --------
        Select rows in data where column values are larger than:

        >>> data.select_by_condition(data["column"] > 2)

        Or select rows in the data based on multiple conditions:

        >>> data.select_by_condition((data["column1"] > 2) & (data["column2] < 1))

        """
        if invert:
            selected = self[~condition]
        else:
            selected = self[condition]
        return self.__class__(selected, self.has_inclined)

    def get_cumulative_thickness(self):  # pragma: no cover
        raise NotImplementedError()

    def get_layer_top(self):  # pragma: no cover
        raise NotImplementedError()

    def to_pyvista_cylinders(self):  # pragma: no cover
        raise NotImplementedError()

    def to_pyvista_grid(self):  # pragma: no cover
        raise NotImplementedError()

    def to_datafusiontools(self):  # pragma: no cover
        raise NotImplementedError()


class Collection(AbstractCollection):
    """
    A collection combines header and data and ensures that they remain aligned when
    applying methods.

    Parameters
    ----------
    header : :class:`~geost.base.PointHeader` or :class:`~geost.base.LineHeader`
        Instance of a header class corresponding to the data.
    data : :class:`~geost.base.LayeredData` or :class:`~geost.base.DiscreteData`
        Instance of a data object corresponding to the header.
    """

    def __init__(
        self,
        header: HeaderObject | None,
        data: DataObject | None,
    ):
        # TODO inheritance refactor: check init types and default values.
        self.header = header
        self.data = data

    def __repr__(self):
        if "_header" in self.__dict__:
            return f"{self.__class__.__name__}:\n# header = {self.n_points}"
        else:
            return f"{self.__class__.__name__}:\n<EMPTY COLLECTION>"

    def __len__(self):
        return len(self.header)

    @property
    def header(self):
        """
        The collection's header.
        """
        return self._header

    @property
    def data(self):
        """
        The collection's data.
        """
        return self._data

    @property
    def n_points(self):
        """
        Number of objects in the collection.
        """
        return len(self)

    @property
    def horizontal_reference(self):
        """
        Coordinate reference system represented by an instance of pyproj.crs.CRS.
        """
        return self.header.horizontal_reference

    @property
    def vertical_reference(self):
        """
        Vertical datum represented by an instance of pyproj.crs.CRS.
        """
        return self.header.vertical_reference

    @property
    def has_inclined(self):
        """
        Boolean indicating whether there are inclined objects within the collection
        """
        return self.data.has_inclined

    @header.setter
    def header(self, header):
        if isinstance(header, LineHeader | PointHeader):
            self._header = header
        elif "_header" in self.__dict__.keys() and isinstance(header, gpd.GeoDataFrame):
            self._header = self._header.__class__(header, self.vertical_reference)
        self.check_header_to_data_alignment()

    @data.setter
    def data(self, data):
        if isinstance(data, LayeredData | DiscreteData):
            self._data = data
        elif "_data" in self.__dict__.keys() and isinstance(data, pd.DataFrame):
            self._data = self._data.__class__(data, self.has_inclined)
        self.check_header_to_data_alignment()

    def _clone_with_attrs(self, new_header, new_data):
        """
        Create a deep copy of the current object with new header and data attributes.
        This is used to return a new instance of the collection in methods that modify
        the number of collection objects through e.g. selection and slicing methods.
        Using this method over the self.__class__ constructor ensures that the new
        object only has the header and data attributes updated, while keeping all other
        attributes intact.

        Parameters
        ----------
        new_header : Any Header object
            The new header to be assigned to the cloned collection.
        new_data : Any Data object
            The new data to be assigned to the cloned collection.

        Returns
        -------
        new_collection : New instance of self
            A deep copy of the current object with updated header and data attributes.
        """
        new_collection = deepcopy(self)
        new_collection._header, new_collection._data = new_header, new_data
        return new_collection

    def add_header_column_to_data(self, column_name: str):  # No change
        """
        Add a column from the header to the data table. Useful if you e.g. add some data
        to the header table, but would like to add this to each layer (row in the data
        table) as well.

        Parameters
        ----------
        column_name : str
            Name of the column in the header table to add.
        """
        self.data.df = pd.merge(self.data.df, self.header[["nr", column_name]], on="nr")

    def get(self, selection_values: str | Iterable, column: str = "nr"):
        """
        Get a subset of a collection through a string or iterable of object id(s).
        Optionally uses a different column than "nr" (the column with object ids).

        Parameters
        ----------
        selection_values : str | Iterable
            Values to select
        column : str, optional
            In which column of the header to look for selection values, by default "nr"

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        Examples
        --------
        self.get(["obj1", "obj2"]) will return a collection with only these objects.

        Suppose we have a collection of boreholes that we have joined with geological
        map units using the method
        :meth:`~geost.base.PointDataCollection.get_area_labels`. We have added this data
        to the header table in the column 'geological_unit'. Using:

        self.get(["unit1", "unit2"], column="geological_unit")

        will return a :class:`~geost.base.BoreholeCollection` with all boreholes
        that are located in "unit1" and "unit2" geological map areas.
        """
        selected_header = self.header.get(selection_values, column)
        selected_data = self.data.select_by_values(column, selection_values)

        return self.__class__(selected_header, selected_data)

    def change_horizontal_reference(self, to_epsg: str | int | CRS):
        """
        Change the horizontal reference (i.e. coordinate reference system, crs) of the
        collection to the given target crs.

        Parameters
        ----------
        to_epsg : str | int | CRS
            EPSG of the target crs. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input().

        Examples
        --------
        To change the collection's current horizontal reference to WGS 84 UTM zone 31N:

        >>> self.change_horizontal_reference(32631)

        This would be the same as:

        >>> self.change_horizontal_reference("epsg:32631")

        As Pyproj is very flexible, you can even use the CRS's full official name:

        >>> self.change_horizontal_reference("WGS 84 / UTM zone 31N")
        """
        transformer = horizontal_reference_transformer(
            self.horizontal_reference, to_epsg
        )
        self.data.df[["x", "y"]] = self.data.df[["x", "y"]].astype(float)
        self.data.df.loc[:, "x"], self.data.df.loc[:, "y"] = transformer.transform(
            self.data["x"], self.data["y"]
        )
        if self.data.has_inclined:
            self.data.df[["x_bot", "y_bot"]] = self.data.df[["x_bot", "y_bot"]].astype(
                float
            )
            self.data.df.loc[:, "x_bot"], self.data.df.loc[:, "y_bot"] = (
                transformer.transform(self.data["x_bot"], self.data["y_bot"])
            )

        self.header.change_horizontal_reference(to_epsg)

    def change_vertical_reference(self, to_epsg: str | int | CRS):
        """
        Change the vertical reference of the collection object's surface levels

        Parameters
        ----------
        to_epsg : str | int | CRS
            EPSG of the target vertical datum. Takes anything that can be interpreted by
            pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum.

            Some often-used vertical datums are:
            - NAP : 5709
            - MSL NL depth : 9288
            - LAT NL depth : 9287
            - Ostend height : 5710

            See epsg.io for more.

        Examples
        --------
        To change the header's current vertical reference to NAP:

        >>> self.change_vertical_reference(5709)

        This would be the same as:

        >>> self.change_vertical_reference("epsg:5709")

        As the Pyproj constructors are very flexible, you can even use the CRS's full
        official name instead of an EPSG number. E.g. for changing to NAP and the
        Belgian Ostend height vertical datums repsectively, you can use:

        >>> self.change_vertical_reference("NAP")
        >>> self.change_vertical_reference("Ostend height")
        """
        transformer = vertical_reference_transformer(
            self.horizontal_reference, self.vertical_reference, to_epsg
        )
        self.data.df[["surface", "end"]] = self.data.df[["surface", "end"]].astype(
            float
        )
        _, _, new_surface = transformer.transform(
            self.data["x"], self.data["y"], self.data["surface"]
        )
        _, _, new_end = transformer.transform(
            self.data["x"], self.data["y"], self.data["end"]
        )
        self.data.df.loc[:, "surface"] = new_surface
        self.data.df.loc[:, "end"] = new_end
        self.header.change_vertical_reference(to_epsg)

    def reset_header(self):
        """
        Refresh the header based on the loaded data in case the header got messed up.
        """
        self.header = self.data.to_header(
            self.horizontal_reference, self.vertical_reference
        )

    def check_header_to_data_alignment(self):
        """
        Two-way check to warn of any misalignment between the header and data
        attributes. Two way, i.e. if header includes more objects than in the data and
        if the data includes more unique objects that listed in the header.

        This check is performed everytime the object is instantiated AND if any change
        is made to either the header or data attributes (see their respective setters).
        """
        warning_given = False
        if hasattr(self, "_header") and hasattr(self, "_data"):
            if any(~self.header["nr"].isin(self.data["nr"].unique())):
                warnings.warn(
                    "Header covers more/other objects than present in the data table. "
                    "consider running the method 'reset_header' to update the header.",
                    category=AlignmentWarning,
                )
                warning_given = True

            if not set(self.data["nr"].unique()).issubset(set(self.header["nr"])):
                warnings.warn(
                    "Header does not cover all unique objects in data. "
                    "consider running the method 'reset_header' to update the header.",
                    category=AlignmentWarning,
                )
                warning_given = True

            if config.validation.AUTO_ALIGN and warning_given:
                self.reset_header()
                print(
                    "\nNOTE: Header has been reset to align with data because AUTO_ALIGN"
                    " is enabled in the GeoST configuration.",
                )

    def select_within_bbox(
        self,
        xmin: Coordinate,
        ymin: Coordinate,
        xmax: Coordinate,
        ymax: Coordinate,
        invert: bool = False,
    ):
        """
        Make a selection of the collection based on a bounding box.

        Parameters
        ----------
        xmin : float | int
            Minimum x-coordinate of the bounding box.
        ymin : float | int
            Minimum y-coordinate of the bounding box.
        xmax : float | int
            Maximum x-coordinate of the bounding box.
        ymax : float | int
            Maximum y-coordinate of the bounding box.
        invert : bool, optional
            Invert the selection, so select all objects outside of the
            bounding box in this case, by default False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.
        """
        header_selected = self.header.select_within_bbox(
            xmin, ymin, xmax, ymax, invert=invert
        )
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_with_points(
        self,
        points: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int,
        invert: bool = False,
    ):
        """
        Make a selection of the collection based on point geometries.

        Parameters
        ----------
        points : str | Path | gpd.GeoDataFrame | GeometryType
            Any type of point geometries that can be used for the selection: GeoDataFrame
            containing points or filepath to a shapefile like file, or Shapely Point,
            MultiPoint or list containing Point objects.
        buffer : float | int
            Buffer distance for selection geometries.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        """
        header_selected = self.header.select_with_points(points, buffer, invert=invert)
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_with_lines(
        self,
        lines: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int,
        invert: bool = False,
    ):
        """
        Make a selection of the collection based on line geometries.

        Parameters
        ----------
        lines : str | Path | gpd.GeoDataFrame | GeometryType
            Any type of line geometries that can be used for the selection: GeoDataFrame
            containing lines or filepath to a shapefile like file, or Shapely LineString,
            MultiLineString or list containing LineString objects.
        buffer : float | int
            Buffer distance for selection geometries.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        """
        header_selected = self.header.select_with_lines(lines, buffer, invert=invert)
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_within_polygons(
        self,
        polygons: str | Path | gpd.GeoDataFrame | GeometryType,
        buffer: float | int = 0,
        invert: bool = False,
    ):
        """
        Make a selection of the collection based on polygon geometries.

        Parameters
        ----------
        polygons : str | Path | gpd.GeoDataFrame | PolygonGeometries
            Any type of polygon geometries that can be used for the selection: GeoDataFrame
            containing polygons or filepath to a shapefile like file, or Shapely Polygon,
            MultiPolygon or list containing Polygon objects.
        buffer : float | int, optional
            Optional buffer distance around the polygon selection geometries, by default
            0.
        invert : bool, optional
            Invert the selection, by default False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        """
        header_selected = self.header.select_within_polygons(
            polygons, buffer=buffer, invert=invert
        )
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_by_depth(
        self,
        top_min: float = None,
        top_max: float = None,
        end_min: float = None,
        end_max: float = None,
    ):
        """
        Select data from depth constraints. If a keyword argument is not given it will
        not be considered. e.g. if you need only boreholes that go deeper than -500 m
        use only end_max = -500.

        Parameters
        ----------
        top_min : float, optional
            Minimum elevation of the borehole/cpt top, by default None.
        top_max : float, optional
            Maximum elevation of the borehole/cpt top, by default None.
        end_min : float, optional
            Minimum elevation of the borehole/cpt end, by default None.
        end_max : float, optional
            Maximumelevation of the borehole/cpt end, by default None.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.
        """
        header_selected = self.header.select_by_depth(
            top_min=top_min, top_max=top_max, end_min=end_min, end_max=end_max
        )
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_by_length(self, min_length: float = None, max_length: float = None):
        """
        Select data from length constraints: e.g. all boreholes between 50 and 150 m
        long. If a keyword argument is not given it will not be considered.

        Parameters
        ----------
        min_length : float, optional
            Minimum length of borehole/cpt, by default None.
        max_length : float, optional
            Maximum length of borehole/cpt, by default None.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.
        """
        header_selected = self.header.select_by_length(
            min_length=min_length, max_length=max_length
        )
        data_selected = self.data.select_by_values("nr", header_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_by_values(
        self, column: str, selection_values: str | Iterable, how: str = "or"
    ):
        """
        Select data based on the presence of given values in a given column. Can be used
        for example to select boreholes that contain peat in the lithology column.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data to use when looking for
            values.
        selection_values : str | Iterable
            Value or values to look for in the column.
        how : str, optional
            Either "and" or "or". "and" requires all selection values to be present in
            column for selection. "or" will select the core if any one of the
            selection_values are found in the column. Default is "and".

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        Examples
        --------
        To select boreholes where both clay ("K") and peat ("V") are present at the same
        time, use "and" as a selection method:

        >>> boreholes.select_by_values("lith", ["V", "K"], how="and")

        To select boreholes that can have one, or both lithologies, use or as the selection
        method:

        >>> boreholes.select_by_values("lith", ["V", "K"], how="and")

        """
        data_selected = self.data.select_by_values(column, selection_values, how=how)
        header_selected = self.header.get(data_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def slice_depth_interval(
        self,
        upper_boundary: float | int = None,
        lower_boundary: float | int = None,
        relative_to_vertical_reference: bool = False,
        update_layer_boundaries: bool = True,
    ):
        """
        Slice data based on given upper and lower boundaries. This returns a new object
        containing only the sliced data.

        Parameters
        ----------
        upper_boundary : float | int, optional
            Every layer that starts above this is removed. The default is None.
        lower_boundary : float | int, optional
            Every layer that starts below this is removed. The default is None.
        relative_to_vertical_reference : bool, optional
            If True, the slicing is done with respect to any kind of vertical reference
            plane (e.g. "NAP", "TAW"). If False, the slice is done with respect to depth
            below the surface. The default is False.
        update_layer_boundaries : bool, optional
            If True, the layer boundaries in the sliced data are updated according to the
            upper and lower boundaries used with the slice. If False, the original layer
            boundaries are kept in the sliced object. The default is False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        Examples
        --------
        Usage depends on whether the slicing is done with respect to depth below the
        surface or to a vertical reference plane.

        For example, select layers in data that are between 2 and 3 meters below the
        surface:

        >>> data.slice_depth_interval(2, 3)

        By default, the method updates the layer boundaries in sliced object according to
        the upper and lower boundaries. To suppress this behaviour use:

        >>> data.slice_depth_interval(2, 3, update_layer_boundaries=False)

        Slicing can also be done with respect to a vertical reference plane like "NAP".
        For example, to select layers in data that are between -3 and -5 m NAP, use:

        >>> data.slice_depth_interval(-3, -5, relative_to_vertical_reference=True)

        """
        data_selected = self.data.slice_depth_interval(
            upper_boundary=upper_boundary,
            lower_boundary=lower_boundary,
            relative_to_vertical_reference=relative_to_vertical_reference,
            update_layer_boundaries=update_layer_boundaries,
        )
        header_selected = self.header.get(data_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def slice_by_values(
        self, column: str, selection_values: str | Iterable, invert: bool = False
    ):
        """
        Slice rows from data based on matching condition. E.g. only return rows with
        a certain lithology in the collection object.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data to use when looking for
            values.
        selection_values : str | Iterable
            Values to look for in the column.
        invert : bool, optional
            If True, invert the slicing action, so remove layers with selected values
            instead of keeping them. The default is False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        Examples
        --------
        Return only rows in borehole data contain sand ("Z") as lithology:

        >>> boreholes.slice_by_values("lith", "Z")

        If you want all the rows that may contain everything but sand, use the "invert"
        option:

        >>> boreholes.slice_by_values("lith", "Z", invert=True)

        """
        data_selected = self.data.slice_by_values(
            column, selection_values, invert=invert
        )
        header_selected = self.header.get(data_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def select_by_condition(self, condition: Any, invert: bool = False):
        """
        Select from collection.data using a manual condition that results in a boolean
        mask. Returns the rows in the data where the 'condition' evaluates to True.

        Parameters
        ----------
        condition : list, pd.Series or array like
            Boolean array like object with locations at which the values will be
            preserved, dtype must be 'bool' and the length must correspond with the
            length of the data.
        invert : bool, optional
            If True, the selection is inverted so rows that evaluate to False will be
            returned. The default is False.

        Returns
        -------
        :class:`~geost.base.LayeredData`
            New instance containing only the data objects selected by this method.

        Examples
        --------
        Select rows in data that contain a specific value:

        >>> data.select_by_condition(data["lith"]=="V")

        Select rows in the data that contain a specific (part of) string or strings:

        >>> boreholes.select_by_condition(boreholes["column"].str.contains("foo|bar"))

        Select rows in data where column values are larger than:

        >>> data.select_by_condition(data["column"] > 2)

        Or select rows in the data based on multiple conditions:

        >>> data.select_by_condition((data["column1"] > 2) & (data["column2] < 1))

        """
        data_selected = self.data.select_by_condition(condition, invert)
        header_selected = self.header.get(data_selected["nr"].unique())
        return self._clone_with_attrs(header_selected, data_selected)

    def get_area_labels(
        self,
        polygon_gdf: str | Path | gpd.GeoDataFrame,
        column_name: str | Iterable,
        include_in_header=False,
    ) -> pd.DataFrame:
        """
        Find in which area (polygons) the point data locations fall. e.g. to determine
        in which geomorphological unit points are located.

        Parameters
        ----------
        polygon_gdf : str | Path | gpd.GeoDataFrame
            GeoDataFrame with polygons.
        column_name : str | Iterable
            The column name to find the labels in. Given as a string or iterable of
            strings in case you'd like to find multiple labels.
        include_in_header : bool, optional
            Whether to add the acquired data to the header table or not, by default
            False.

        Returns
        -------
        pd.DataFrame
            Borehole ids and the polygon label they are in. If include_in_header = True,
            a column containing the generated data will be added inplace to the header.
        """
        result = self.header.get_area_labels(
            polygon_gdf, column_name, include_in_header=include_in_header
        )
        return result

    def to_geoparquet(self, outfile: str | Path, **kwargs):
        """
        Write header data to geoparquet. You can use the resulting file to display
        borehole locations in GIS for instance. Please note that Geoparquet is supported
        by GDAL >= 3.5. For Qgis this means QGis >= 3.26

        Parameters
        ----------
        file : str | Path
            Path to shapefile to be written.
        **kwargs
            gpd.GeoDataFrame.to_parquet kwargs. See relevant Pandas documentation.

        """
        self.header.to_geoparquet(outfile, **kwargs)

    def to_shape(self, outfile: str | Path, **kwargs):
        """
        Write header data to shapefile. You can use the resulting file to display
        borehole locations in GIS for instance.

        Parameters
        ----------
        file : str | Path
            Path to shapefile to be written.
        **kwargs
            gpd.GeoDataFrame.to_file kwargs. See relevant GeoPandas documentation.

        """
        self.header.to_shape(outfile, **kwargs)

    def to_geopackage(self, outfile: str | Path, metadata: dict[str, str] = None):
        """
        Write header and data to a geopackage. You can use the resulting file to display
        borehole locations in GIS for instance.

        Parameters
        ----------
        file : str | Path
            Path to geopackage to be written.
        metadata : dict[str, str], default None
            Optional metadata to be stored in the file. Keys and values must be strings.
            The default is None.

        """
        utils._to_geopackage(
            self.header.gdf,
            outfile,
            "Header",
            layer="header",
            driver="GPKG",
            index=False,
            metadata=metadata,
        )
        utils._to_geopackage(
            gpd.GeoDataFrame(self.data.df),
            outfile,
            "Data",
            layer="data",
            driver="GPKG",
            index=False,
            metadata=metadata,
        )

    def to_parquet(self, outfile: str | Path, data_table: bool = True, **kwargs):
        """
        Export the data or header table to a parquet file. By default the data table is
        exported.

        Parameters
        ----------
        file : str | Path
            Path to parquet file to be written.
        data_table : bool, optional
            If True, the data table is exported. If False, the header table is exported.
        **kwargs
            pd.DataFrame.to_parquet kwargs. See relevant Pandas documentation.

        Examples
        --------
        Export the data table:
        >>> collection.to_parquet("example.parquet")

        Export the header table:
        >>> collection.to_parquet("example.parquet", data_table=False)

        """
        if data_table:
            self.data.to_parquet(outfile, **kwargs)
        else:
            self.header.to_parquet(outfile, **kwargs)

    def to_csv(self, outfile: str | Path, data_table: bool = True, **kwargs):
        """
        Export the data or header table to a csv file. By default the data table is
        exported.

        Parameters
        ----------
        file : str | Path
            Path to csv file to be written.
        data_table : bool, optional
            If True, the data table is exported. If False, the header table is exported.
        **kwargs
            pd.DataFrame.to_csv kwargs. See relevant Pandas documentation.

        Examples
        --------
        Export the data table:
        >>> collection.to_parquet("example.csv")

        Export the header table:
        >>> collection.to_parquet("example.csv", data_table=False)

        """
        if data_table:
            self.data.to_csv(outfile, **kwargs)
        else:
            self.header.to_csv(outfile, **kwargs)

    def to_pickle(self, outfile: str | Path, **kwargs):
        """
        Export the complete Collection to a pickle file.

        Parameters
        ----------
        file : str | Path
            Path to pickle file to be written.
        **kwargs
            pd.to_pickle kwargs. See relevant Pandas documentation.

        Examples
        --------
        >>> collection.to_pickle("example.pkl")

        """
        utils.save_pickle(self, outfile, **kwargs)

    def to_pyvista_cylinders(
        self,
        displayed_variables: str | List[str],
        radius: float = 1,
        vertical_factor: float = 1.0,
        relative_to_vertical_reference: bool = True,
    ):
        """
        Create a Pyvista MultiBlock object of cylinder-shaped geometries to represent
        boreholes. Although cylinders are prettier when visualized, they are quite costly
        to render in large numbers. Consider using
        :meth:`~geost.base.Collection.to_pyvista_grid` instead for large datasets.

        Parameters
        ----------
        displayed_variables : str | List[str]
            Name or names of data columns to include for visualisation. Can be columns that
            contain an array of floats, ints and strings.
        radius : float, optional
            Radius of the cylinders in m in the MultiBlock. The default is 1.
        vertical_factor : float, optional
            Factor to correct vertical scale. For example, when layer boundaries are given
            in cm, use 0.01 to convert to m. The default is 1.0, so no correction is applied.
            It is not recommended to use this for vertical exaggeration, use viewer functionality
            for that instead.
        relative_to_vertical_reference : bool, optional
            If True, the depth of the objects in the vtm file will be with respect to a
            reference plane (e.g. "NAP", "TAW"). If False, the depth will be with respect
            to 0.0. The default is True.

        Returns
        -------
        pyvista.MultiBlock
            A composite class holding the data which can be iterated over.

        """
        return self.data.to_pyvista_cylinders(
            displayed_variables, radius, vertical_factor, relative_to_vertical_reference
        )

    def to_pyvista_grid(
        self,
        displayed_variables: str | List[str],
        radius: float = 1.0,
    ):
        """
        Create a Pyvista UnstructuredGrid object to represent boreholes. This is more efficient
        than :meth:`~geost.base.Collection.to_pyvista_cylinders` for large datasets, but
        less visually appealing.

        Parameters
        ----------
        displayed_variables : str | List[str]
            Name or names of data columns to include for visualisation. Can be columns that
            contain an array of floats, ints and strings.
        radius : float, optional
            Radius of the cylinders in m in the MultiBlock. The default is 1.

        Returns
        -------
        pyvista.UniformGrid
            A grid class holding the data which can be iterated over.

        """
        return self.data.to_pyvista_grid(displayed_variables, radius)

    def to_datafusiontools(
        self,
        columns: List[str],
        outfile: str | Path = None,
        encode: bool = False,
        relative_to_vertical_reference: bool = True,
    ):
        """
        Export all data to the core "Data" class of Deltares DataFusionTools. Returns
        a list of "Data" objects, one for each data object that is exported. This list
        can directly be used within DataFusionTools. If out_file is given, the list of
        Data objects is saved to a pickle file.

        For DataFusionTools visit:
        https://bitbucket.org/DeltaresGEO/datafusiontools/src/master/

        Parameters
        ----------
        columns : List[str]
            Which columns in the data to include for the export. These will become variables
            in the DataFusionTools "Data" class.
        outfile : str | Path, optional
            If a path to outfile is given, the data is written to a pickle file.
        encode : bool, default True
            If True, categorical data columns are encoded to additional binary columns
            (all possible values become a seperate feature that is 0 or 1). The default is
            False. Warning: if there is a large number of possible categories, many columns
            with categorical data or both, the export process may become slow and may consume
            a large amount memory. Please consider carefully which categorical data columns
            need to be included.
        relative_to_vertical_reference : bool, optional
            If True, the depth of all data objects will converted to a depth with respect to
            a reference plane (e.g. "NAP", "Ostend height"). If False, the depth will be
            kept as original in the "top" and "bottom" columns which is in meter below
            the surface. The default is True.

        Returns
        -------
        List[Data]
            List containing the DataFusionTools Data objects.

        """
        return self.data.to_datafusiontools(
            columns, outfile, encode, relative_to_vertical_reference
        )


class BoreholeCollection(Collection):
    """
    A collection combines header and borehole data and ensures that they remain aligned
    when applying methods.

    Parameters
    ----------
    header : :class:`~geost.base.PointHeader`
        Instance of a header class corresponding to the data.
    data : :class:`~geost.base.LayeredData`
        Instance of a data object corresponding to the header.
    """

    def add_grainsize_data(self, sample_data: pd.DataFrame):
        """
        Add grain size data to the borehole collection.

        Parameters
        ----------
        sample_data : pd.DataFrame
            DataFrame containing sample data with a column "nr" that contains borehole ids.

        """
        # safe_validate(DataSchemas.grainsize_data, sample_data, inplace=True)
        # sample_data["nr"].unique()
        # warnings.warn(
        #     "Header covers more/other objects than present in the data table, "
        #     "consider running the method 'reset_header' to update the header.",
        #     AlignmentWarning,
        # )
        # self.sample_data = sample_data
        raise NotImplementedError

    def get_cumulative_thickness(
        self, column: str, values: str | List[str], include_in_header: bool = False
    ):
        """
        Get the cumulative thickness of layers where a column contains a specified search
        value or values.

        Parameters
        ----------
        column : str
            Name of column that must contain the search value or values.
        values : str | List[str]
            Search value or values in the column to find the cumulative thickness for.
        include_in_header : bool, optional
            If True, include the result in the header table of the collection. In this
            case, the method does not return a DataFrame. The default is False.

        Returns
        -------
        pd.DataFrame
            Borehole ids and cumulative thickness of selected layers if the "include_in_header"
            option is set to False.

        Examples
        --------
        Get the cumulative thickness of the layers with lithology "K" in the column "lith"
        use:

        >>> boreholes.get_cumulative_thickness("lith", "K")

        Or get the cumulative thickness for multiple selection values. In this case, a
        Pandas DataFrame is returned with a column per selection value containing the
        cumulative thicknesses:

        >>> boreholes.get_cumulative_thickness("lith", ["K", "Z"])

        To include the result in the header object of the collection, use the
        "include_in_header" option:

        >>> boreholes.get_cumulative_thickness("lith", ["K"], include_in_header=True)

        """
        cum_thickness = self.data.get_cumulative_thickness(column, values)
        cum_thickness.columns = cum_thickness.columns.astype(str)

        if include_in_header:
            columns = [c + "_thickness" for c in cum_thickness.columns]
            self.header.gdf.drop(
                columns=columns,
                errors="ignore",
                inplace=True,
            )
            self.header = self.header.gdf.merge(
                cum_thickness.add_suffix("_thickness"), on="nr", how="left"
            )
            self.header[columns] = self.header[columns].fillna(0)
        else:
            return cum_thickness

    def get_layer_top(
        self,
        column: str,
        values: str | List[str],
        min_thickness: float = 0,
        min_depth: float = 0,
        include_in_header: bool = False,
    ):
        """
        Find the depth at which a specified layer first occurs, starting at min_depth
        and looking downwards until the first layer of min_thickness is found of the
        specified layer.

        Parameters
        ----------
        column : str
            Name of column that contains categorical data.
        values : str | List[str]
            Value or values of entries in the column that you want to find top of.
        min_thickness : float, optional
            Minimal thickness of the layer to be considered. The default is 0.
        min_depth : float, optional
            Minimal depth of the layer to be considered. The default is 0.
        include_in_header : bool, optional
            If True, include the result in the header table of the collection. In this
            case, the method does not return a DataFrame. The default is False.

        Returns
        -------
        pd.DataFrame
            Borehole ids and top levels of selected layers in meters below the surface.

        Examples
        --------
        Get the top depth of layers in boreholes where the lithology in the "lith" column
        is sand ("Z"):

        >>> boreholes.get_layer_top("lith", "Z")

        To include the result in the header object of the collection, use the
        "include_in_header" option:

        >>> boreholes.get_layer_top("lith", "Z", include_in_header=True)

        """
        tops = self.data.get_layer_top(
            column, values, min_thickness=min_thickness, min_depth=min_depth
        )

        if include_in_header:
            self.header.gdf.drop(
                columns=[c + "_top" for c in tops.columns],
                errors="ignore",
                inplace=True,
            )
            self.header = self.header.gdf.merge(
                tops.add_suffix("_top"), on="nr", how="left"
            )
        else:
            return tops

    def to_qgis3d(
        self,
        outfile: str | Path,
        relative_to_vertical_reference: bool = True,
        **kwargs,
    ):
        """
        Write data to geopackage file that can be directly loaded in the Qgis2threejs
        plugin. Works only for layered (borehole) data.

        Parameters
        ----------
        outfile : str | Path
            Path to geopackage file to be written.
        relative_to_vertical_reference : bool, optional
            If True, the depth of all data objects will converted to a depth with respect to
            a reference plane (e.g. "NAP", "TAW"). If False, the depth will be kept as original
            in the "top" and "bottom" columns which is in meter below the surface. The default
            is True.

        **kwargs
            geopandas.GeodataFrame.to_file kwargs. See relevant Geopandas documentation.

        """
        self.data.to_qgis3d(
            outfile,
            relative_to_vertical_reference,
            crs=self.horizontal_reference,
            **kwargs,
        )

    def to_kingdom(
        self,
        outfile: str | Path,
        tdstart: int = 1,
        vw: float = 1500.0,
        vs: float = 1600.0,
    ):
        """
        Write data to 2 csv files: interval data and time-depth chart,
            for import in Kingdom seismic interpretation software.

        Parameters
        ----------
        out_file : str | Path
            Path to csv file to be written.
        tdstart : int
            startindex for TDchart, default is 1
        vw : float
            sound velocity in water in m/s, default is 1500 m/s
        vs : float
            sound velocity in sediment in m/s, default is 1600 m/s
        """
        self.data.to_kingdom(outfile, tdstart, vw, vs)


class CptCollection(Collection):
    def slice_depth_interval(
        self,
        upper_boundary: float | int = None,
        lower_boundary: float | int = None,
        relative_to_vertical_reference: bool = False,
    ):
        """
        Slice data based on given upper and lower boundaries. This returns a new object
        containing only the sliced data.

        Parameters
        ----------
        upper_boundary : float | int, optional
            Every layer that starts above this is removed. The default is None.
        lower_boundary : float | int, optional
            Every layer that starts below this is removed. The default is None.
        relative_to_vertical_reference : bool, optional
            If True, the slicing is done with respect to any kind of vertical reference
            plane (e.g. "NAP", "TAW"). If False, the slice is done with respect to depth
            below the surface. The default is False.

        Returns
        -------
        New instance of the current object.
            New instance of the current object containing only the selection resulting
            from application of this method. e.g. if you are calling this method from a
            Collection, you will get an instance of a Collection back.

        Examples
        --------
        Usage depends on whether the slicing is done with respect to depth below the
        surface or to a vertical reference plane.

        For example, select layers in data that are between 2 and 3 meters below the
        surface:

        >>> data.slice_depth_interval(2, 3)

        By default, the method updates the layer boundaries in sliced object according to
        the upper and lower boundaries. To suppress this behaviour use:

        >>> data.slice_depth_interval(2, 3, update_layer_boundaries=False)

        Slicing can also be done with respect to a vertical reference plane like "NAP".
        For example, to select layers in data that are between -3 and -5 m NAP, use:

        >>> data.slice_depth_interval(-3, -5, relative_to_vertical_reference=True)

        """
        data_selected = self.data.slice_depth_interval(
            upper_boundary=upper_boundary,
            lower_boundary=lower_boundary,
            relative_to_vertical_reference=relative_to_vertical_reference,
        )
        collection_selected = data_selected.to_collection()
        return collection_selected

    def get_cumulative_thickness(self):  # pragma: no cover
        raise NotImplementedError()

    def get_layer_top(self):  # pragma: no cover
        raise NotImplementedError()


class LogCollection(Collection):  # pragma: no cover
    def get_cumulative_thickness(self):
        raise NotImplementedError()

    def get_layer_top(self):
        raise NotImplementedError()



from pathlib import Path
from typing import Any, Callable

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from geost.io import xml
from geost.io.xml import schemas


@pytest.fixture
def xml_string(testdatadir: Path):
    with open(testdatadir / r"xml/bhrgt_bro.xml", "rb") as file:
        string = file.read()
    return string


@pytest.mark.parametrize(
    "file, reader, expected_end_nap",
    [
        ("bhrgt_bro.xml", xml.read_bhrgt, -6.91),
        ("bhrg_bro.xml", xml.read_bhrg, -2.31),
        ("bhrp_bro.xml", xml.read_bhrp, -1.29),
        ("cpt_bro.xml", xml.read_cpt, -6.48),
        ("sfr_bro.xml", xml.read_sfr, -0.73),
    ],
    ids=["bhrgt", "bhrg", "bhrp", "cpt", "sfr"],
)
def test_read(testdatadir: Path, file: str, reader: Callable, expected_end_nap: float):
    bro_xml = testdatadir / r"xml" / file
    header, data = xml.read(bro_xml, reader)
    assert isinstance(header, pd.DataFrame)
    assert isinstance(data, pd.DataFrame)
    expected_columns_present = ["nr", "x", "y", "surface", "end"]
    assert all(c in data.columns for c in expected_columns_present)

    if "top" in data.columns and "bottom" in data.columns:
        assert data["top"].dtype == data["bottom"].dtype == float

    assert_array_almost_equal(header["end"], expected_end_nap)
    assert_array_almost_equal(data["end"], expected_end_nap)


class TestBhrgt:
    @pytest.fixture
    def custom_bhrgt_schema(self):
        return {
            "payload_root": "dispatchDocument",
            "bro_id": {"xpath": "brocom:broId"},
            "date": {"xpath": "reportHistory/reportStartDate/brocom:date"},
        }

    @pytest.mark.unittest
    def test_read_bhrgt(self, testdatadir: Path):
        bro_xml = testdatadir / r"xml/bhrgt_bro.xml"

        data = xml.read_bhrgt(bro_xml)

        assert isinstance(data, dict)
        assert data["nr"] == "BHR000000336600"

        with pytest.raises(SyntaxError, match="Invalid xml schema"):
            xml.read_bhrgt(bro_xml, company="Wiertsema")

        with pytest.raises(
            ValueError, match="No predefined schema for 'UnknownCompany'"
        ):
            xml.read_bhrgt(bro_xml, company="UnknownCompany")

    @pytest.mark.unittest
    def test_read_bhrgt_with_custom_schema(
        self, testdatadir: Path, custom_bhrgt_schema: dict[str, Any]
    ):
        data = xml.read_bhrgt(
            testdatadir / r"xml/bhrgt_bro.xml", schema=custom_bhrgt_schema
        )
        assert isinstance(data, dict)
        assert data == {"bro_id": "BHR000000336600", "date": "2020-04-15"}

    @pytest.mark.unittest
    def test_read_bhrgt_from_string(self, xml_string: bytes):
        data = xml.read_bhrgt(xml_string)
        assert isinstance(data, dict)
        assert data["nr"] == "BHR000000336600"

    @pytest.mark.unittest
    def test_read_bhrgt_with_custom_schema_no_payload_root(self, testdatadir: Path):
        """
        Test reading an xml string without a payload root. This can lead to unexpected results
        as the XML root can contain multiple elements with additional information besides the
        main data element.

        """
        schema = {"bro_id": {"xpath": "BHR_GT_O/brocom:broId"}}
        with pytest.warns(UserWarning, match="Multiple payloads found in XML"):
            data = xml.read_bhrgt(
                testdatadir / r"xml/bhrgt_bro.xml", schema=schema, read_all=False
            )
            assert isinstance(data, dict)

        data = xml.read_bhrgt(
            testdatadir / r"xml/bhrgt_bro.xml", schema=schema, read_all=True
        )
        assert isinstance(data, list)
        assert data[0] == {"bro_id": None}
        assert data[1] == {"bro_id": None}
        assert data[2] == {"bro_id": None}
        assert data[3] == {"bro_id": "BHR000000336600"}

    @pytest.mark.unittest
    def test_read_bhrgt_bro(self, testdatadir: Path):
        data = xml.read_bhrgt(testdatadir / r"xml/bhrgt_bro.xml")

        assert isinstance(data, dict)
        assert data["nr"] == "BHR000000336600"
        assert data["location"] == (132781.327, 448031.1)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 0.09
        assert data["vertical_datum"] == "NAP"
        assert data["groundwater_level"] == 1.6
        assert data["end"] == 7.0
        assert data["data"] == {
            "upperBoundary": [
                "0.00",
                "1.00",
                "1.70",
                "2.00",
                "2.40",
                "3.40",
                "4.20",
                "5.00",
                "6.00",
            ],
            "lowerBoundary": [
                "1.00",
                "1.70",
                "2.00",
                "2.40",
                "3.40",
                "4.20",
                "5.00",
                "6.00",
                "7.00",
            ],
            "geotechnicalSoilName": [
                "sterkZandigeKleiMetGrind",
                "zwakGrindigZand",
                "zwakZandigeKleiMetGrind",
                "zand",
                "detritus",
                "detritus",
                "zwakZandigeKlei",
                "zand",
                "zand",
            ],
        }

    @pytest.mark.unittest
    def test_read_bhrgt_wiertsema(self, testdatadir: Path):
        data = xml.read_bhrgt(
            testdatadir / r"xml/bhrgt_wiertsema.xml", company="Wiertsema"
        )

        assert isinstance(data, dict)
        assert data["nr"] == "_87078_HB001"
        assert data["location"] == (182243.9, 335073.8)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 37.74
        assert data["vertical_datum"] == "NAP"
        assert data["groundwater_level"] == 1.1
        assert data["end"] == 2.2
        assert data["data"] == {
            "upperBoundary": ["0.00", "0.10", "0.40", "0.70", "1.70"],
            "lowerBoundary": ["0.10", "0.40", "0.70", "1.70", "2.20"],
            "geotechnicalSoilName": [
                "zwakZandigSilt",
                "zwakZandigGrind",
                "klei",
                "zand",
                "zand",
            ],
        }


class TestBhrg:
    @pytest.mark.unittest
    def test_read_bhrg(self, testdatadir: Path):
        bro_xml = testdatadir / r"xml/bhrg_bro.xml"

        data = xml.read_bhrg(bro_xml)
        assert isinstance(data, dict)

        with pytest.raises(
            ValueError, match="No predefined schema for 'UnknownCompany'"
        ):
            xml.read_bhrg(bro_xml, company="UnknownCompany")

        invalid_schema = schemas.bhrgt.get("Wiertsema", None)
        with pytest.raises(SyntaxError, match="Invalid xml schema"):
            xml.read_bhrg(bro_xml, schema=invalid_schema)

    @pytest.mark.unittest
    def test_read_bhrg_bro(self, testdatadir: Path):
        data = xml.read_bhrg(testdatadir / r"xml/bhrg_bro.xml")

        assert isinstance(data, dict)
        assert data["nr"] == "BHR000000396406"
        assert data["location"] == (126149.0, 452162.0)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 0.69
        assert data["vertical_datum"] == "NAP"
        assert data["end"] == 3.0
        assert data["data"] == {
            "upperBoundary": ["0.000", "0.250", "1.600", "2.000", "2.500"],
            "lowerBoundary": ["0.250", "1.600", "2.000", "2.500", "3.000"],
            "soilNameNEN5104": [
                "zwakZandigeKlei",
                "sterkSiltigeKlei",
                "zwakZandigeKlei",
                "sterkZandigeKlei",
                "zwakSiltigZand",
            ],
        }


class TestCpt:
    @pytest.mark.unittest
    def test_read_cpt(self, testdatadir: Path):
        data = xml.read_cpt(testdatadir / r"xml/cpt_bro.xml")

        assert isinstance(data, dict)

        with pytest.raises(
            ValueError, match="No predefined schema for 'UnknownCompany'"
        ):
            xml.read_cpt(testdatadir / r"xml/cpt_bro.xml", company="UnknownCompany")

        invalid_schema = schemas.bhrgt.get("Wiertsema", None)
        with pytest.raises(SyntaxError, match="Invalid xml schema"):
            xml.read_cpt(testdatadir / r"xml/cpt_bro.xml", schema=invalid_schema)

    @pytest.mark.unittest
    def test_read_cpt_bro(self, testdatadir: Path):
        data = xml.read_cpt(testdatadir / r"xml/cpt_bro.xml")

        assert isinstance(data, dict)
        assert data["nr"] == "CPT000000155283"
        assert data["location"] == (132782.52, 448030.34)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 0.09
        assert data["vertical_datum"] == "NAP"
        assert data["predrilled_depth"] == 0.5
        assert data["end"] == 6.57
        assert_array_equal(
            list(data["data"].keys()),
            [
                "penetrationlength",
                "depth",
                "elapsedtime",
                "coneresistance",
                "correctedconeresistance",
                "netconeresistance",
                "magneticfieldstrengthx",
                "magneticfieldstrengthy",
                "magneticfieldstrengthz",
                "magneticfieldstrengthtotal",
                "electricalconductivity",
                "inclinationew",
                "inclinationns",
                "inclinationx",
                "inclinationy",
                "inclinationresultant",
                "magneticinclination",
                "magneticdeclination",
                "localfriction",
                "poreratio",
                "temperature",
                "porepressureu1",
                "porepressureu2",
                "porepressureu3",
                "frictionratio",
            ],
        )
        original_missing_value = -999999
        processed_data = np.array(list(data["data"].values()))
        assert not np.any(processed_data == original_missing_value)


class TestBhrp:
    @pytest.mark.unittest
    def test_read_bhrp(self, testdatadir: Path):
        data = xml.read_bhrp(testdatadir / r"xml/bhrp_bro.xml")
        assert isinstance(data, dict)

        with pytest.raises(
            ValueError, match="No predefined schema for 'UnknownCompany'"
        ):
            xml.read_bhrp(testdatadir / r"xml/bhrp_bro.xml", company="UnknownCompany")

        invalid_schema = schemas.bhrgt.get("Wiertsema", None)
        with pytest.raises(SyntaxError, match="Invalid xml schema"):
            xml.read_bhrp(testdatadir / r"xml/bhrp_bro.xml", schema=invalid_schema)

    @pytest.mark.unittest
    def test_read_bhrp_bro(self, testdatadir: Path):
        data = xml.read_bhrp(testdatadir / r"xml/bhrp_bro.xml")

        assert isinstance(data, dict)
        assert data["nr"] == "BHR000000108193"
        assert data["location"] == (129491.0, 452255.0)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 0.21
        assert data["vertical_datum"] == "NAP"
        assert data["begin_depth"] == 0.0
        assert data["end"] == 1.5
        assert data["ghg"] == 0.55
        assert data["glg"] == 1.6
        assert data["landuse"] == "graslandBlijvend"
        assert data["data"] == {
            "upperBoundary": ["0.000", "0.600", "0.900", "1.200", "1.300", "1.400"],
            "lowerBoundary": ["0.600", "0.900", "1.200", "1.300", "1.400", "1.500"],
            "standardSoilName": [
                "sterkSiltigeKlei",
                "sterkSiltigeKlei",
                "sterkSiltigeKlei",
                "uiterstSiltigeKlei",
                "zwakSiltigZand",
                "zwakSiltigZand",
            ],
        }


class TestSfr:
    @pytest.mark.unittest
    def test_read_sfr(self, testdatadir: Path):
        data = xml.read_sfr(testdatadir / r"xml/sfr_bro.xml")
        assert isinstance(data, dict)

        with pytest.raises(
            ValueError, match="No predefined schema for 'UnknownCompany'"
        ):
            xml.read_sfr(testdatadir / r"xml/sfr_bro.xml", company="UnknownCompany")

        invalid_schema = schemas.bhrgt.get("Wiertsema", None)
        with pytest.raises(SyntaxError, match="Invalid xml schema"):
            xml.read_sfr(testdatadir / r"xml/sfr_bro.xml", schema=invalid_schema)

    @pytest.mark.unittest
    def test_read_sfr_bro(self, testdatadir: Path):
        data = xml.read_sfr(testdatadir / r"xml/sfr_bro.xml")

        assert isinstance(data, dict)
        assert data["nr"] == "SFR000000000687"
        assert data["location"] == (132250.0, 451075.0)
        assert data["crs"] == "urn:ogc:def:crs:EPSG::28992"
        assert data["surface"] == 0.87
        assert data["vertical_datum"] == "NAP"
        assert data["landuse"] == "grasland"
        assert data["outcrop_type"] == "profielkuil"
        assert data["end"] == 1.6
        assert data["data"] == {
            "upperBoundary": [
                "0.000",
                "0.070",
                "0.250",
                "0.550",
                "0.830",
                "1.200",
                "1.400",
            ],
            "lowerBoundary": [
                "0.070",
                "0.250",
                "0.550",
                "0.830",
                "1.200",
                "1.400",
                "1.600",
            ],
            "soilNameNEN5104": [
                "sterkSiltigeKlei",
                "sterkSiltigeKlei",
                "matigSiltigeKlei",
                "sterkSiltigeKlei",
                "uiterstSiltigeKlei",
                "matigZandigeKlei",
                "sterkZandigeKlei",
            ],
        }



from geost.io.xml import resolvers

BRO = {
    "payload_root": "dispatchDocument",
    "nr": {"xpath": "brocom:broId"},
    "location": {
        "xpath": "deliveredLocation/bhrcommon:location/gml:pos",
        "resolver": resolvers.parse_coordinates,
        "el-attr": "text",
    },
    "crs": {
        "xpath": "deliveredLocation/bhrcommon:location",
        "resolver": resolvers.parse_crs,
    },
    "surface": {
        "xpath": "deliveredVerticalPosition/bhrcommon:offset",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "vertical_datum": {
        "xpath": "deliveredVerticalPosition/bhrcommon:verticalDatum",
        "el-attr": "text",
    },
    "begin_depth": {
        "xpath": "boring/bhrcommon:boredTrajectory/bhrcommon:beginDepth",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "end": {
        "xpath": "boring/bhrcommon:boredTrajectory/bhrcommon:endDepth",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "ghg": {
        "xpath": "boreholeSampleDescription/bhrcommon:result/bhrcommon:meanHighestGroundwaterLevel",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "glg": {
        "xpath": "siteCharacteristic/bhrcommon:meanLowestGroundwaterTable",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "landuse": {"xpath": "siteCharacteristic/bhrcommon:landUse"},
    "data": {
        "xpath": "boreholeSampleDescription/bhrcommon:result",
        "resolver": resolvers.process_bhrp_data,
        "layer-attributes": ["upperBoundary", "lowerBoundary", "standardSoilName"],
    },
}

schemas = {"BRO": BRO}



import logging
import re
from pathlib import Path
from typing import NamedTuple, Union

import numpy as np
import pandas as pd
from shapely.geometry import Point


class ColumnInfo(NamedTuple):
    value: str
    unit: str
    name: str
    standard: bool


class CptMeasurementVar(NamedTuple):
    value: Union[int, float, str]
    unit: str
    quantity: str
    reserved: bool


#
COLUMN_DEFS_DATA_BLOCK_CPT = {
    1: ColumnInfo("length", "m", "penetration length", True),
    2: ColumnInfo("qc", "MPa", "measured cone resistance", True),
    3: ColumnInfo("fs", "MPa", "friction resistance", True),
    4: ColumnInfo("rf", "%", "friction number", True),
    5: ColumnInfo("u1", "MPa", "pore pressure u1", True),
    6: ColumnInfo("u2", "MPa", "pore pressure u2", True),
    7: ColumnInfo("u3", "MPa", "pore pressure u3", True),
    8: ColumnInfo("inclination_res", "degrees", "inclination (resultant)", True),
    9: ColumnInfo("inclination_ns", "degrees", "inclination (North-South)", True),
    10: ColumnInfo("inclination_ew", "degrees", "inclination (East-West)", True),
    11: ColumnInfo(
        "corrected_depth", "m", "corrected depth, below fixed surface", True
    ),  # noqa: E501
    12: ColumnInfo("time", "s", "time", True),
    13: ColumnInfo("qt", "MPa", "corrected cone resistance", True),
    14: ColumnInfo("qn", "MPa", "net cone resistance", True),
    15: ColumnInfo("Bq", "", "pore ratio", True),
    16: ColumnInfo("Nm", "", "cone resistance number", True),
    17: ColumnInfo("gamma", "kN/m3", "weight per unit volume", True),
    18: ColumnInfo("u0", "MPa", "in situ, initial pore pressure", True),
    19: ColumnInfo("sigma", "MPa", "total vertical soil pressure", True),
    20: ColumnInfo("sigma_eff", "MPa", "effective vertical soil pressure", True),
    21: ColumnInfo("inclination_x", "degrees", "Inclination in X direction", True),
    22: ColumnInfo("inclination_y", "degrees", "Inclination in Y direction", True),
    23: ColumnInfo("ec", "S/m", "Electric conductivity", True),
    24: ColumnInfo("Bx", "nT", "magnetic field strength in X direction", True),
    25: ColumnInfo("By", "nT", "magnetic field strength in Y direction", True),
    26: ColumnInfo("Bz", "nT", "magnetic field strength in Z direction", True),
    # reserved for future use
    27: ColumnInfo("", "degrees", "magnetic inclination", True),
    # reserved for future use
    28: ColumnInfo("", "degrees", "magnetic inclination", True),
}


RESERVED_MEASUREMENTVARS_CPT = {
    1: CptMeasurementVar(1000, "mm2", "nom. surface area cone tip", True),
    2: CptMeasurementVar(15000, "mm2", "nom. surface area friction sleeve", True),
    3: CptMeasurementVar(None, "", "net surface area quotient of cone tip", True),
    4: CptMeasurementVar(
        None, "", "net surface area quotient of friction sleeve", True
    ),  # noqa: E501
    5: CptMeasurementVar(
        100, "mm", "distance of cone to centre of friction sleeve", True
    ),  # noqa: E501
    6: CptMeasurementVar(None, "", "friction present", True),
    7: CptMeasurementVar(None, "", "PPT u1 present", True),
    8: CptMeasurementVar(None, "", "PPT u2 present", True),
    9: CptMeasurementVar(None, "", "PPT u3 present", True),
    10: CptMeasurementVar(None, "", "inclination measurement present", True),
    11: CptMeasurementVar(None, "", "use of back-flow compensator", True),
    12: CptMeasurementVar(None, "", "type of cone penetration test", True),
    13: CptMeasurementVar(None, "m", "pre-excavated depth", True),
    14: CptMeasurementVar(None, "m", "groundwater level", True),
    15: CptMeasurementVar(None, "m", "water depth (for offshore)", True),
    16: CptMeasurementVar(None, "m", "end depth of penetration test", True),
    17: CptMeasurementVar(None, "", "stop criteria", True),
    # 18: CptMeasurementVar(None, '', 'for future use', True),
    # 19: CptMeasurementVar(None, '', 'for future use', True),
    20: CptMeasurementVar(None, "MPa", "zero measurement cone before", True),
    21: CptMeasurementVar(None, "MPa", "zero measurement cone after", True),
    22: CptMeasurementVar(None, "MPa", "zero measurement friction before", True),
    23: CptMeasurementVar(None, "MPa", "zero measurement friction after", True),
    24: CptMeasurementVar(None, "MPa", "zero measurement PPT u1 before", True),
    25: CptMeasurementVar(None, "MPa", "zero measurement PPT u1 after", True),
    26: CptMeasurementVar(None, "MPa", "zero measurement PPT u2 before", True),
    27: CptMeasurementVar(None, "MPa", "zero measurement PPT u2 after", True),
    28: CptMeasurementVar(None, "MPa", "zero measurement PPT u3 before", True),
    29: CptMeasurementVar(None, "MPa", "zero measurement PPT u3 after", True),
    30: CptMeasurementVar(None, "degrees", "zero measurement inclination before", True),
    31: CptMeasurementVar(None, "degrees", "zero measurement inclination after", True),
    32: CptMeasurementVar(
        None, "degrees", "zero measurement inclination NS before", True
    ),  # noqa: E501
    33: CptMeasurementVar(
        None, "degrees", "zero measurement inclination NS after", True
    ),  # noqa: E501
    34: CptMeasurementVar(
        None, "degrees", "zero measurement inclination EW before", True
    ),  # noqa: E501
    35: CptMeasurementVar(
        None, "degrees", "zero measurement inclination EW after", True
    ),  # noqa: E501
    # 36: CptMeasurementVar(None, '', 'for future use', True),
    # 37: CptMeasurementVar(None, '', 'for future use', True),
    # 38: CptMeasurementVar(None, '', 'for future use', True),
    # 39: CptMeasurementVar(None, '', 'for future use', True),
    # 40: CptMeasurementVar(None, '', 'for future use', True),
    41: CptMeasurementVar(None, "km", "mileage", True),
    42: CptMeasurementVar(
        None, "degrees", "Orientation between X axis inclination and North", True
    ),  # noqa: E501
}


GEF_CPT_REFERENCE_LEVELS = {
    "00000": "own reference level",
    "00001": "Low Low Water Spring",
    "31000": "NAP",
    "32000": "Ostend Level",
    "32001": "TAW",
    "49000": "Normall Null",
}


class CptGefFile:  # TODO: Break parser down in HeaderParser and DataParser and returning a CPT object
    """
    Parse a .gef file of a CPT sounding for its content to retrieve the data and
    relevant information about the measurements.

    Parameters
    ----------
    path : str, Path, optional
        Path to the gef file to parse. If None, an empty class instance is returned.
    sep : str, optional
        Column separator character of the gef file to use when the separator is not
        specified within the gef file header.

    """

    def __init__(self, path: str | Path = None, sep: str = " "):
        self.path = path
        self._header = None
        self._data = None

        self.nr = None
        self.x = None
        self.y = None
        self.z = None
        self.enddepth = None

        # mandatory gef header attributes
        self.gefid = None
        self.ncolumns = None
        self.columninfo = dict()
        self.companyid = None
        self.filedate = None
        self.fileowner = None
        self.lastscan = None
        self.procedurecode = None  # mandatory if gefid is 1, 0, 0
        self.reportcode = None  # this or procedurecode is mandatory if gefid > 1, 1, 0
        self.projectid = None
        self.measurementtext = dict()

        # Additional gef header attributes
        self.columnvoid = dict()
        self.columnminmax = None
        self.columnseparator = sep  # default separator used if not in gef file header
        self.dataformat = None
        self.measurementvars = dict()
        self.recordseparator = None
        self.reportdataformat = None
        self.specimenvars = None
        self.startdate = None
        self.starttime = None
        self.coord_system = None
        self.reference_system = None

        if path:
            self.__open_file(path)

    def __repr__(self):
        return f"{self.__class__.__name__}(nr={self.nr})"

    def __open_file(self, path):
        with open(path, "r", encoding="utf-8", errors="ignore") as f:
            text = f.read()
            end_header = re.search(r"(?P<eoh>#EOH[=\s+]+)", text).group("eoh")

            self._header, self._data = text.split(end_header)

        self.parse_header()
        self.parse_data()

        self.get_enddepth()

    @property
    def df(self):
        if not hasattr(self, "_df"):
            self.to_df()
        return self._df

    @property
    def header(self):
        header = pd.Series(
            [self.nr, self.x, self.y, self.z, self.enddepth, self.point],
            index=["nr", "x", "y", "z", "enddepth", "geometry"],
        )
        return header

    @property
    def columns(self):
        columns = [f"{c.value}" for c in self.columninfo.values()]
        return columns

    @property
    def point(self):
        return Point(self.x, self.y)

    @staticmethod
    def to_zero_indexed(idx: str):
        """
        Update an index from the gef file (is 1-indexed) to a 0-indexed Python
        index.

        """
        return int(idx) - 1

    def parse_header(self):
        header = self._header.splitlines()

        for line in header:
            keyword = re.search(r"([#\s]*([A-Z]+)\s*=)\s*", line)

            try:
                keyword_method = keyword.group(2).lower()
            except AttributeError:
                continue

            __method = f"_parse_{keyword_method}"
            if hasattr(self, __method):
                line = line.replace(keyword.group(0), "")
                # remove unnecessary whitespace and string quotes
                line = re.sub(r'["\']|\s\s+', "", line)
                self.__call_header_method(__method, line)

    @staticmethod
    def __split_line(line):
        return [l.strip() for l in line.split(",")]  # noqa: E741

    def parse_data(self):
        """
        Parse datablock of the gef file.

        """
        data = self._data

        if not self.recordseparator:
            r = "!"
        else:
            r = self.recordseparator

        data = data.replace(self.columnseparator, ",")
        data = [self.__split_line(d.rstrip(f",{r}")) for d in data.splitlines()]
        self._data = data

    def to_df(self):
        """
        Create a Pandas DataFrame from the gef datablock.

        Returns
        -------
        None.

        """
        df = pd.DataFrame(self._data)
        df.dropna(inplace=True)
        df = df.astype(float)
        df.replace(self.columnvoid, np.nan, inplace=True)
        df.columns = self.columns

        if "rf" not in df.columns:
            try:
                df["rf"] = (df["fs"] / df["qc"]) * 100
            except KeyError:
                logging.warning(
                    f"Missing data in {self.nr}. Data present: {self.columns}"
                )

        if (
            "corrected_depth" in df.columns
        ):  # TODO: implement calc corrected depth from inclination if not in columns  # noqa: E501
            df["depth"] = df["corrected_depth"]
        else:
            df["depth"] = df["length"]

        self._df = df

    def __call_header_method(self, method, line):
        """
        Helper method to call the correct parser method of the class for a specific
        header attribute (e.g. _parse_columninfo).

        Parameters
        ----------
        method : str
            Name of the header attribute.
        line : str
            Line in the header to parse.

        """
        return getattr(self, method)(line)

    def _parse_gefid(self, line):
        self.gefid = line

    def _parse_column(self, line):
        self.ncolumns = int(line)

    def _parse_columninfo(self, line: str):
        idx, unit, value, number = self.__split_line(line)
        idx = self.to_zero_indexed(idx)
        info = COLUMN_DEFS_DATA_BLOCK_CPT.get(int(number), "empty")

        if info == "empty":
            logging.warning(f"Unknown information in datablock of {self.path}")
            info = ColumnInfo(value, unit, value, False)

        self.columninfo.update({idx: info})

    def _parse_companyid(self, line: str):
        self.companyid = line

    def _parse_filedate(self, line: str):
        pass

    def _parse_fileowner(self, line: str):
        self.fileowner = line

    def _parse_lastscan(self, line: str):
        self.lastscan = int(line)

    def _parse_procedurecode(self, line: str):
        self.procedurecode = line

    def _parse_reportcode(self, line: str):
        self.reportcode = line

    def _parse_projectid(self, line: str):
        self.projectid = line

    def _parse_testid(self, line: str):
        self.nr = line

    # TODO: check how to fix if zid occurs in header more than once
    def _parse_zid(self, line: str):
        zid = self.__split_line(line)
        if len(zid) == 2:
            reference_system = zid[0]
            self.z = float(zid[1])
            self.delta_z = None
        elif len(zid) == 3:
            reference_system = zid[0]
            self.z = float(zid[1])
            self.delta_z = float(zid[2])
        else:
            logging.warning(
                f"Unclear information in #ZID of {self.path}. "
                "Check zid attribute manually."
            )
            self.zid = zid

        self.reference_system = GEF_CPT_REFERENCE_LEVELS[reference_system]

    # TODO: add correct parsing of reserved measurementtexts
    def _parse_measurementtext(self, line: str):
        text = self.__split_line(line)
        nr, info = int(text[0]), text[1:]
        self.measurementtext.update({nr: info})

    def _parse_xyid(self, line: str):
        xyid = self.__split_line(line)

        if len(xyid) == 3:
            self.coord_system = xyid[0]
            self.x = float(xyid[1])
            self.y = float(xyid[2])

        elif len(xyid) == 5:
            self.coord_system = xyid[0]
            self.x = float(xyid[1])
            self.y = float(xyid[2])
            self.dx = float(xyid[3])
            self.dy = float(xyid[4])

        else:
            logging.warning(
                f"Unclear information in #XYID of {self.path}. "
                "Check xyid attribute manually."
            )
            self.xyid = xyid

    def _parse_columnvoid(self, line: str):
        idx, value = self.__split_line(line)
        idx = self.to_zero_indexed(idx)
        self.columnvoid.update({idx: float(value)})

    def _parse_columnminmax(self, line: str):
        pass

    def _parse_columnseparator(self, line: str):
        self.columnseparator = line

    def _parse_dataformat(self, line: str):
        pass

    def _parse_measurementvar(self, line: str):
        num, val, unit, quantity = self.__split_line(line)

        num = int(num)
        val = safe_float(val)

        _mv = RESERVED_MEASUREMENTVARS_CPT.get(num, "empty")

        if _mv == "empty":
            mvar = CptMeasurementVar(val, unit, quantity, False)
        else:
            if val:
                mvar = CptMeasurementVar(val, _mv.unit, _mv.quantity, True)
            else:
                mvar = _mv

        self.measurementvars.update({num: mvar})

    def _parse_recordseparator(self, line: str):
        self.recordseparator = line

    def _parse_reportdataformat(self, line: str):
        pass

    def _parse_specimenvar(self, line: str):
        pass

    def _parse_startdate(self, line: str):
        pass

    def _parse_starttime(self, line: str):
        pass

    def get_enddepth(self):
        enddepth = self.measurementvars.get(16)
        if enddepth:
            d = enddepth.value
        else:
            d = self.df["length"].max()

        self.enddepth = self.z - d


def safe_float(number):
    try:
        return float(number)
    except ValueError:
        return None



from pathlib import Path

import geopandas as gpd
import pytest
from numpy.testing import assert_array_equal

from geost.bro import BroCptGeopackage
from geost.io import Geopackage


class TestBroCptGeopackage:
    @pytest.mark.unittest
    def test_from_geopackage(self, bro_cpt_gpkg: Path):
        bro_cpt = BroCptGeopackage.from_geopackage(bro_cpt_gpkg)

        assert isinstance(bro_cpt.gdf, gpd.GeoDataFrame)
        assert isinstance(bro_cpt.db, Geopackage)
        assert bro_cpt.gdf.index.name == "fid"
        assert_array_equal(
            bro_cpt.gdf.columns,
            [
                "bro_id",
                "quality_regime",
                "delivery_accountable_party",
                "delivery_context",
                "survey_purpose",
                "research_report_date",
                "cpt_standard",
                "additional_investigation_performed",
                "applied_transformation",
                "geometry",
            ],
        )

        with pytest.raises(ValueError):
            BroCptGeopackage.from_geopackage(
                bro_cpt_gpkg, layer="geotechnical_cpt_survey"
            )

        bro_cpt = BroCptGeopackage.from_geopackage(bro_cpt_gpkg, fid_as_index=False)
        assert bro_cpt.gdf.index.name is None

    @pytest.mark.unittest
    def test_select_location_info(self, bro_cpt_gpkg: Path):
        bro_cpt = BroCptGeopackage.from_geopackage(bro_cpt_gpkg)
        with pytest.raises(NotImplementedError):
            bro_cpt.select_location_info()



from typing import Literal

import pytest
from numpy.testing import assert_array_equal

from geost.bro import GeoTop
from geost.bro.bro_geotop import StratGeotop


class TestGeoTop:
    @pytest.fixture
    def test_area(self):
        """
        xmin, ymin, xmax, ymax bounding box of a test area.
        """
        return (115_000, 500_000, 115_500, 500_500)

    @pytest.mark.unittest
    def test_lazy_load_from_opendap(self):
        geotop = GeoTop.from_opendap(lazy=True)
        assert geotop.ds.chunks == {
            "y": (75, 684, 684, 684, 684),
            "x": (643, 643, 643, 643, 74),
            "z": (76, 76, 76, 76, 9),
        }

    @pytest.mark.unittest
    def test_from_opendap(
        self,
        test_area: tuple[
            Literal[115000], Literal[500000], Literal[115500], Literal[500500]
        ],
    ):
        geotop = GeoTop.from_opendap(bbox=test_area)
        assert isinstance(geotop, GeoTop)
        assert geotop.resolution == (100, 100, 0.5)
        assert geotop["strat"].dims == ("y", "x", "z")
        assert geotop.crs == 28992
        assert_array_equal(
            geotop["x"], [114_950, 115_050, 115_150, 115_250, 115_350, 115_450, 115_550]
        )
        assert_array_equal(
            geotop["y"], [500_550, 500_450, 500_350, 500_250, 500_150, 500_050, 499_950]
        )
        assert_array_equal(geotop["z"][:3], [-49.75, -49.25, -48.75])


class TestStratGeotop:
    @pytest.mark.unittest
    def test_select_units(self):
        sel = StratGeotop.select_units(["AAOP", "OEC", "BXDE", "ANAWA"])
        assert_array_equal(
            sel,
            [
                StratGeotop.holocene.OEC,
                StratGeotop.channel.ANAWA,
                StratGeotop.older.BXDE,
                StratGeotop.antropogenic.AAOP,
            ],
        )
        assert_array_equal(sel, [1070, 6010, 3040, 1000])

        sel = StratGeotop.select_units("AEC")
        assert_array_equal(sel, [StratGeotop.channel.AEC])
        assert_array_equal(sel, [6000])

    @pytest.mark.unittest
    def test_select_values(self):
        sel = StratGeotop.select_values([1070, 6010, 3040, 1000])
        assert_array_equal(
            sel,
            [
                StratGeotop.holocene.OEC,
                StratGeotop.channel.ANAWA,
                StratGeotop.older.BXDE,
                StratGeotop.antropogenic.AAOP,
            ],
        )
        assert_array_equal(sel, [1070, 6010, 3040, 1000])

        sel = StratGeotop.select_values(6000)
        assert_array_equal(sel, [StratGeotop.channel.AEC])
        assert_array_equal(sel, [6000])

    @pytest.mark.unittest
    def test_select_units_empty(self):
        sel = StratGeotop.select_units("foo")
        assert not sel

    @pytest.mark.unittest
    def test_select_values_empty(self):
        sel = StratGeotop.select_values(-9999)
        assert not sel

    @pytest.mark.unittest
    def test_to_dict(self):
        d = StratGeotop.to_dict(key="unit")
        assert d == {
            "NIGR": 1010,
            "NINB": 1045,
            "NASC": 1020,
            "ONAWA": 1030,
            "NAZA": 1040,
            "NAWAZU": 1048,
            "NAWAAL": 1049,
            "NAWA": 1050,
            "BHEC": 1060,
            "OEC": 1070,
            "NAWOBE": 1080,
            "KK1": 1085,
            "NIFL": 1089,
            "NIHO": 1090,
            "NAZA2": 1095,
            "NAWO": 1100,
            "NWNZ": 1110,
            "NAWOVE": 1120,
            "KK2": 1125,
            "NIBA": 1130,
            "NA": 2000,
            "EC": 2010,
            "NI": 2020,
            "KK": 2030,
            "AEC": 6000,
            "ABEOM": 6005,
            "ANAWA": 6010,
            "ANAWO": 6020,
            "BEC": 6100,
            "BNAWA": 6110,
            "BNAWO": 6120,
            "CEC": 6200,
            "CNAWA": 6210,
            "CNAWO": 6220,
            "DEC": 6300,
            "DNAWA": 6310,
            "DNAWO": 6320,
            "EEC": 6400,
            "ENAWA": 6410,
            "ENAWO": 6420,
            "BXKO": 3000,
            "BXSI": 3010,
            "BXSI1": 3011,
            "BXWI": 3020,
            "BXSI2": 3012,
            "BXWIKO": 3025,
            "BXWISIKO": 3030,
            "BXDE": 3040,
            "BXDEKO": 3045,
            "BXSC": 3050,
            "BXLM": 3060,
            "BXBS": 3090,
            "BX": 3100,
            "KRWY": 4000,
            "KRBXDE": 4010,
            "KRZU": 4020,
            "KROE": 4030,
            "KRTW": 4040,
            "KR": 4050,
            "BEOM": 4055,
            "BEWY": 4060,
            "BERO": 4070,
            "BE": 4080,
            "KW1": 4085,
            "KW": 4090,
            "WB": 4100,
            "EE": 4110,
            "EEWB": 4120,
            "DR": 5000,
            "DRGI": 5010,
            "GE": 5020,
            "DN": 5030,
            "URTY": 5040,
            "PE": 5050,
            "UR": 5060,
            "ST": 5070,
            "AP": 5080,
            "SY": 5090,
            "PZ": 5100,
            "WA": 5110,
            "PZWA": 5120,
            "MS": 5130,
            "KI": 5140,
            "OO": 5150,
            "IE": 5160,
            "VI": 5170,
            "BR": 5180,
            "VE": 5185,
            "RUBO": 5190,
            "RU": 5200,
            "TOZEWA": 5210,
            "TOGO": 5220,
            "TO": 5230,
            "DOAS": 5240,
            "DOIE": 5250,
            "DO": 5260,
            "LA": 5270,
            "HT": 5280,
            "HO": 5290,
            "MT": 5300,
            "GU": 5310,
            "VA": 5320,
            "AK": 5330,
            "AAOP": 1000,
            "AAES": 1005,
        }
        assert len(d) == (
            len(StratGeotop.holocene)
            + len(StratGeotop.channel)
            + len(StratGeotop.older)
            + len(StratGeotop.antropogenic)
        )



from enum import IntEnum, StrEnum
from typing import Iterable, Literal


class UnitEnum(IntEnum):
    """
    Extension of IntEnum to be used for classified values of stratigraphic units and
    lithologies in models to allow easy selections.

    """

    @classmethod
    def select_units(cls, units: str | Iterable[str]):
        """
        Select units by name.

        Parameters
        ----------
        units : str | Iterable[str]
            Name as string or array_like object of strings with names to select.

        Returns
        -------
        List[enum]
            List of the selected units.

        """
        units = cls.cast_to_list(units)
        return [u for u in cls if u.name in units]

    @classmethod
    def select_values(cls, values: int | float | Iterable[int | float]):
        """
        Select units by value.

        Parameters
        ----------
        units : int | float | Iterable[int | float]
            Number or array_like object of numbers for units to select.

        Returns
        -------
        List[enum]
            List of the selected units.

        """
        values = cls.cast_to_list(values)
        return [n for n in cls if n.value in values]

    @classmethod
    def to_dict(cls, key: Literal["unit", "value"] = "value") -> dict:
        """
        Create a mapping dictionary from UnitEnum with "unit": "value" or "value": "unit"
        mapping.

        Parameters
        ----------
        key : Literal["unit", "value"], optional
            Determine what to use for "key" ("unit" or "value"). The default is "value".

        Returns
        -------
        dict

        """
        if key == "value":
            return {unit.value: unit.name for unit in cls}
        elif key == "unit":
            return {unit.name: unit.value for unit in cls}
        else:
            raise ValueError("Input for 'key' is wrong, use: 'unit' or 'value'")

    @staticmethod
    def cast_to_list(value):
        if isinstance(value, (int, float, str)):
            value = [value]
        return value


class VerticalReference(StrEnum):
    DATUM = "Local datum"
    DEPTH = "Depth"
    # Add vertical references



import numpy as np
import pandas as pd
import pytest
from numpy.testing import (
    assert_allclose,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
)

# Local imports
from geost.analysis.combine import (
    _add_to_discrete,
    _add_to_layered,
    add_voxelmodel_variable,
)
from geost.analysis.interpret_cpt import calc_ic, calc_lithology
from geost.analysis.layer_analysis import find_top_sand
from geost.base import Collection


class TestAnalysis:
    @pytest.fixture
    def test_borehole_lith(self):
        return np.array(["K", "V", "K", "Z", "K", "Z", "Z"])

    @pytest.fixture
    def test_borehole_top(self):
        return np.array([0, 1, 1.5, 3, 3.5, 4, 6])

    @pytest.fixture
    def test_borehole_bottom(self):
        return np.array([1, 1.5, 3, 3.5, 4, 6, 10])

    @pytest.fixture
    def test_ic_array(self):
        return np.array([0.1, 0.1, 0.2, 0.3, 0.4, 0.5, 1.0, 2.0, 5.0, 10.0, 20.0])

    @pytest.fixture
    def test_fr_array(self):
        return np.array([9.0, 7.0, 5.5, 5.0, 4.0, 3.0, 3.0, 2.0, 1.5, 1.0, 1.0])

    @pytest.mark.unittest
    def test_top_sand(
        self, test_borehole_lith, test_borehole_top, test_borehole_bottom
    ):
        # After the first encounter of sand (Z) > 40% of the next 1 m must consist of
        # sand as well for this first encounter to be regarded as the top of the sand
        # (and all the above the cover layer)
        top = find_top_sand(
            test_borehole_lith, test_borehole_top, test_borehole_bottom, 0.4, 1
        )
        assert_equal(top, 3.0)

        # After the first encounter of sand (Z) > 60% of the next 1 m must consist of
        # sand as well for this first encounter to be regarded as the top of the sand
        # (and all the above the cover layer)
        top = find_top_sand(
            test_borehole_lith, test_borehole_top, test_borehole_bottom, 0.6, 1
        )
        assert_equal(top, 4.0)

        top = find_top_sand(
            test_borehole_lith, test_borehole_top, test_borehole_bottom, 0.5, 1
        )
        assert_equal(top, 3.0)

        top = find_top_sand(
            test_borehole_lith, test_borehole_top, test_borehole_bottom, 1, 6.5
        )
        assert_equal(top, np.nan)

        top = find_top_sand(
            test_borehole_lith, test_borehole_top, test_borehole_bottom, 0.91, 6.5
        )
        assert_equal(top, 3.0)

    @pytest.mark.unittest
    def test_calc_ic(self, test_ic_array, test_fr_array):
        ic = calc_ic(test_ic_array, test_fr_array)
        assert_allclose(
            ic,
            np.array(
                [
                    4.09490299,
                    4.03801064,
                    3.72631088,
                    3.55524529,
                    3.39779083,
                    3.249435,
                    2.99685177,
                    2.64914385,
                    2.25513149,
                    1.91031411,
                    1.68964223,
                ]
            ),
        )

    @pytest.mark.unittest
    def test_calc_lithology(self, test_ic_array, test_fr_array):
        lith = calc_lithology(
            calc_ic(test_ic_array, test_fr_array), test_ic_array, test_fr_array
        )
        assert_array_equal(
            lith, np.array(["V", "Kh", "Kh", "K", "K", "K", "K", "Kz", "Z", "Z", "Z"])
        )


class TestCombine:
    @pytest.fixture
    def strat_deeper_than_borehole(self):
        return pd.DataFrame(
            {
                "nr": ["A"] * 4,
                "strat": [1, 2, 3, 4],
                "bottom": [-1.5, -2.0, -4.3, -5.1],
            }
        )

    @pytest.fixture
    def strat_deeper_than_cpt(self):
        return pd.DataFrame(
            {
                "nr": ["a"] * 4,
                "strat": [1, 2, 3, 4],
                "bottom": [-1.5, -2.0, -7.5, -10],
            }
        )

    @pytest.mark.unittest
    def test_add_voxelmodel_variable_layered(self, borehole_collection, voxelmodel):
        result = add_voxelmodel_variable(borehole_collection, voxelmodel, "strat")
        assert isinstance(result, Collection)
        assert_array_equal(
            result.data["strat"],
            [
                1.0,
                1.0,
                1.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                2.0,
                1.0,
                1.0,
                2.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                np.nan,
                np.nan,
            ],
        )
        assert_array_almost_equal(
            result.data["top"],
            [
                0.0,
                0.8,
                1.5,
                1.7,
                2.2,
                2.5,
                3.7,
                0.0,
                0.6,
                1.2,
                1.3,
                2.3,
                2.5,
                3.1,
                0.0,
                0.75,
                1.25,
                1.4,
                1.75,
                1.8,
                2.25,
                2.9,
                3.8,
                0.0,
                0.5,
                1.2,
                1.8,
                2.5,
                0.0,
                0.5,
                1.2,
                1.4,
                1.8,
                1.9,
                2.5,
            ],
        )
        assert_array_almost_equal(
            result.data["bottom"],
            [
                0.8,
                1.5,
                1.7,
                2.2,
                2.5,
                3.7,
                4.2,
                0.6,
                1.2,
                1.3,
                2.3,
                2.5,
                3.1,
                3.9,
                0.75,
                1.25,
                1.4,
                1.75,
                1.8,
                2.25,
                2.9,
                3.8,
                5.5,
                0.5,
                1.2,
                1.8,
                2.5,
                3.0,
                0.5,
                1.2,
                1.4,
                1.8,
                1.9,
                2.5,
                3.0,
            ],
        )

    @pytest.mark.unittest
    def test_add_voxelmodel_variable_discrete(self, cpt_collection, voxelmodel):
        result = add_voxelmodel_variable(cpt_collection, voxelmodel, "strat")
        assert isinstance(result, Collection)
        assert_array_equal(
            result.data["strat"],
            [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ],
        )
        assert_array_equal(
            result.data["depth"],
            [
                0.0,
                1.0,
                2.0,
                3.0,
                3.6,
                4.0,
                4.1,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                0.0,
                1.0,
                2.0,
                2.3,
                2.8,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
            ],
        )

    @pytest.mark.unittest
    def test_add_to_layered(self, borehole_data, strat_deeper_than_borehole):
        """
        Single unit test for when the depth of one or more stratigraphic boundaries are
        not overlapping with a borehole at all. These are omitted when the stratigraphy
        is combined.

        """
        borehole = borehole_data.select_by_values("nr", "A")
        result = _add_to_layered(borehole, strat_deeper_than_borehole)
        assert len(result) == 7
        assert_array_equal(result["strat"], [1, 1, 1, 2, 3, 3, 3])
        assert_array_almost_equal(result["bottom"], [0.8, 1.5, 1.7, 2.2, 2.5, 3.7, 4.2])

    @pytest.mark.unittest
    def test_add_to_discrete(self, cpt_data, strat_deeper_than_cpt):
        """
        Single unit test for when the depth of one or more stratigraphic boundaries are
        not overlapping with a cpt at all. These are omitted when the stratigraphy is
        combined.

        """
        cpt = cpt_data.select_by_values("nr", "a")
        result = _add_to_discrete(cpt, strat_deeper_than_cpt)
        assert len(result) == 12
        assert_array_equal(result["strat"], [1, 1, 1, 1, 1, 2, 2, 3, 3, 3, 3, 3])
        assert_array_almost_equal(
            result["depth"], [0, 1, 2, 3, 3.6, 4, 4.1, 5, 6, 7, 8, 9]
        )

    @pytest.mark.unittest
    def test_removes_if_column_is_present(self, borehole_collection, voxelmodel):
        borehole_collection.data["strat"] = 1000
        result = add_voxelmodel_variable(borehole_collection, voxelmodel, "strat")
        assert_array_equal(
            result.data["strat"],
            [
                1.0,
                1.0,
                1.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                2.0,
                1.0,
                1.0,
                2.0,
                2.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                np.nan,
                np.nan,
            ],
        )



# import numpy as np
# import pandas as pd
from pathlib import Path
from typing import Iterable, Union

from geost.io.parsers.gef_parsers import CptGefFile
from geost.utils import get_path_iterable


def _parse_cpt_gef_files(file_or_folder: Union[str, Path]):
    """
    Parse gef files from CPT data into Pandas DataFrames.

    Parameters
    ----------
    file_or_folder : Union[str, Path]
        Gef file or folder with gef files to parse.

    Yields
    ------
    df : pd.DataFrame
        Pandas DataFrame per gef file.

    """
    if isinstance(file_or_folder, (str, Path)):
        files = get_path_iterable(Path(file_or_folder), wildcard="*.gef")

    elif isinstance(file_or_folder, Iterable):
        files = file_or_folder

    for f in files:
        cpt = CptGefFile(f)
        df = cpt.df

        df.insert(0, "nr", cpt.nr)
        df.insert(1, "x", cpt.x)
        df.insert(2, "y", cpt.y)
        df.insert(3, "surface", cpt.z)
        df.insert(4, "end", cpt.enddepth)
        yield df



class ValidationSettings:
    """
    Settings for validation behavior.

    VERBOSE: If True, validation errors will be printed to the console.
    DROP_INVALID: If True, invalid rows will automatically be dropped from (Geo)DataFrames.
    FLAG_INVALID: If True, invalid rows will be flagged in (Geo)DataFrames. Note: only works if DROP_INVALID is False.
    AUTO_ALIGN: If True, collection headers and data tables will automatically be aligned.
    """

    VERBOSE = True
    DROP_INVALID = True
    FLAG_INVALID = False
    AUTO_ALIGN = True


validation = ValidationSettings()



from pathlib import Path
from typing import Any

import numpy as np
import pytest
from lxml import etree
from numpy.testing import assert_array_equal

from geost.io.xml import resolvers


def crs_element():
    el = etree.Element("Point", srsName="urn:ogc:def:crs:EPSG::28992")
    return el


def empty_crs_element():
    el = etree.Element("Point")
    return el


@pytest.fixture
def bhrgt_no_namespace(testdatadir: Path):
    root = etree.parse(testdatadir / r"xml/bhrgt_data_element_no_namespaces.xml")
    return root


@pytest.mark.parametrize(
    "element, expected",
    [
        (crs_element(), "urn:ogc:def:crs:EPSG::28992"),
        (empty_crs_element(), "unknown"),
    ],
)
def test_parse_crs(element: etree.Element, expected: str):
    result = resolvers.parse_crs(element)
    assert result == expected


@pytest.mark.parametrize(
    "coordinates, expected",
    [
        ("1.0 2.0", (1.0, 2.0)),
        ("2.1,2.0", (2.1, 2.0)),
        ("3.1;1.0", (3.1, 1.0)),
        ("1.0   2.3", (1.0, 2.3)),  # Separated by multiple whitespaces
    ],
)
def test_parse_coordinates(coordinates: str, expected: tuple[float, float]):
    result = resolvers.parse_coordinates(coordinates)
    assert result == expected


@pytest.mark.parametrize(
    "invalid_coordinates",
    ["5.1|1.0", "1.0,,2.3", "1.0;;2.3"],
)
def test_parse_coordinates_invalid(invalid_coordinates: str):
    with pytest.raises(ValueError, match="Cannot parse coordinates"):
        resolvers.parse_coordinates(invalid_coordinates)


@pytest.mark.parametrize(
    "value, expected",
    [
        ("1.0", 1.0),
        (1, 1.0),
        (1.0, 1.0),
        (None, None),
        ([1, 2], None),
        ("[1, 2]", None),
    ],
)
def test_safe_float(value: Any, expected: float | None):
    result = resolvers.safe_float(value)
    assert result == expected


@pytest.mark.unittest
def test_process_bhrgt_data(bhrgt_no_namespace):
    """
    Tests resolvers.process_bhrgt_data for an XML element without namespaces. Behaviour
    for "real" BRO-like elements with data in namespaces is tested in test_xml_read
    module.

    """
    result = resolvers.process_bhrgt_data(
        bhrgt_no_namespace.getroot(),
        [
            "upperBoundary",
            "lowerBoundary",
            "geotechnicalSoilName",
            "nonExistingAttribute",  # Make sure non-existing attributes does not raise error
        ],
    )
    assert result == {
        "upperBoundary": ["0.0", "1.2", "2.1"],
        "lowerBoundary": ["1.2", "2.1", "2.5"],
        "geotechnicalSoilName": ["zand", "klei", "silt"],
        "nonExistingAttribute": [None, None, None],
    }

    # Test that at least "upperBoundary", "lowerBoundary" and "geotechnicalSoilName" are
    # retrieved.
    result = resolvers.process_bhrgt_data(bhrgt_no_namespace.getroot(), None)
    assert result == {
        "upperBoundary": ["0.0", "1.2", "2.1"],
        "lowerBoundary": ["1.2", "2.1", "2.5"],
        "geotechnicalSoilName": ["zand", "klei", "silt"],
    }


@pytest.mark.parametrize("element", [crs_element(), None])
def test_safe_get(element):
    result = resolvers.safe_get(element)
    assert result is None


@pytest.mark.unittest
def test_process_bhrp_data(testdatadir: Path):
    xml = etree.parse(testdatadir / r"xml/bhrp_bro.xml").getroot()
    layer_element = xml.find(
        "dispatchDocument/BHR_O/boreholeSampleDescription/bhrcommon:result",
        xml.nsmap,
    )
    result = resolvers.process_bhrp_data(layer_element, None)
    assert result == {
        "upperBoundary": ["0.000", "0.600", "0.900", "1.200", "1.300", "1.400"],
        "lowerBoundary": ["0.600", "0.900", "1.200", "1.300", "1.400", "1.500"],
        "standardSoilName": [
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "uiterstSiltigeKlei",
            "zwakSiltigZand",
            "zwakSiltigZand",
        ],
    }

    result = resolvers.process_bhrp_data(
        layer_element,
        [
            "upperBoundary",
            "lowerBoundary",
            "standardSoilName",
            "containsGravel",
            "clayContent",
            "nonexistingAttribute",  # Make sure non-existing attributes does not raise error
        ],
    )
    assert result == {
        "upperBoundary": ["0.000", "0.600", "0.900", "1.200", "1.300", "1.400"],
        "lowerBoundary": ["0.600", "0.900", "1.200", "1.300", "1.400", "1.500"],
        "standardSoilName": [
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "uiterstSiltigeKlei",
            "zwakSiltigZand",
            "zwakSiltigZand",
        ],
        "containsGravel": ["nee", "nee", "nee", "nee", "nee", "nee"],
        "clayContent": ["30.0", "30.0", "32.0", "20.0", "4.0", "4.0"],
        "nonexistingAttribute": [None, None, None, None, None, None],
    }


@pytest.mark.unittest
def test_process_cpt_data(testdatadir: Path):
    xml = etree.parse(testdatadir / r"xml/cpt_bro.xml").getroot()
    layer_element = xml.find("dispatchDocument/CPT_O/conePenetrometerSurvey", xml.nsmap)
    result = resolvers.process_cpt_data(layer_element)
    assert isinstance(result, dict)
    assert_array_equal(
        list(result.keys()),
        [
            "penetrationlength",
            "depth",
            "elapsedtime",
            "coneresistance",
            "correctedconeresistance",
            "netconeresistance",
            "magneticfieldstrengthx",
            "magneticfieldstrengthy",
            "magneticfieldstrengthz",
            "magneticfieldstrengthtotal",
            "electricalconductivity",
            "inclinationew",
            "inclinationns",
            "inclinationx",
            "inclinationy",
            "inclinationresultant",
            "magneticinclination",
            "magneticdeclination",
            "localfriction",
            "poreratio",
            "temperature",
            "porepressureu1",
            "porepressureu2",
            "porepressureu3",
            "frictionratio",
        ],
    )


@pytest.mark.unittest
def test_strip_tag():
    el = etree.Element("{http://www.opengis.net/gml}Point")
    assert resolvers.strip_tag(el) == "point"

    el = etree.Element("Point")
    assert resolvers.strip_tag(el) == "point"

    el = etree.Element(
        "{http://www.opengis.net/gml}Point", nsmap={"gml": "http://www.opengis.net/gml"}
    )
    assert resolvers.strip_tag(el) == "point"


@pytest.mark.parametrize(
    "text, column_sep, row_sep",
    [
        ("1,2;3,4", ",", ";"),
        ("1 2,3 4", " ", ","),
        ("1;2:3:4", ";", ":"),
        ("1.2,3.4", ".", ","),
        ("1\t2\n3\t4", "\t", "\n"),
    ],
)
def test_textvalues_to_array(text, column_sep, row_sep):
    result = resolvers.textvalues_to_array(text, column_sep, row_sep)
    assert isinstance(result, np.ndarray)
    assert result.dtype == float
    assert_array_equal(result, [[1.0, 2.0], [3.0, 4.0]])


@pytest.mark.unittest
def test_process_bhrg_data(testdatadir: Path):
    xml = etree.parse(testdatadir / r"xml/bhrg_bro.xml").getroot()
    layer_element = xml.find(
        "dispatchDocument/BHR_G_O/boreholeSampleDescription/bhrgcom:BoreholeSampleDescription/bhrgcom:descriptiveBoreholeLog/bhrgcom:DescriptiveBoreholeLog",
        xml.nsmap,
    )
    result = resolvers.process_bhrg_data(layer_element, None)
    assert result == {
        "upperBoundary": ["0.000", "0.250", "1.600", "2.000", "2.500"],
        "lowerBoundary": ["0.250", "1.600", "2.000", "2.500", "3.000"],
        "soilNameNEN5104": [
            "zwakZandigeKlei",
            "sterkSiltigeKlei",
            "zwakZandigeKlei",
            "sterkZandigeKlei",
            "zwakSiltigZand",
        ],
    }

    result = resolvers.process_bhrg_data(
        layer_element,
        [
            "upperBoundary",
            "lowerBoundary",
            "soilNameNEN5104",
            "carbonateContentClass",
            "constituentType",
            "nonexistingAttribute",  # Make sure non-existing attributes does not raise error
        ],
    )
    assert result == {
        "upperBoundary": ["0.000", "0.250", "1.600", "2.000", "2.500"],
        "lowerBoundary": ["0.250", "1.600", "2.000", "2.500", "3.000"],
        "soilNameNEN5104": [
            "zwakZandigeKlei",
            "sterkSiltigeKlei",
            "zwakZandigeKlei",
            "sterkZandigeKlei",
            "zwakSiltigZand",
        ],
        "carbonateContentClass": [
            "onbekend",
            "onbekend",
            "onbekend",
            "onbekend",
            "onbekend",
        ],
        "constituentType": ["puin", None, None, None, None],
        "nonexistingAttribute": [None, None, None, None, None],
    }


@pytest.mark.unittest
def test_process_bhrg_data(testdatadir: Path):
    xml = etree.parse(testdatadir / r"xml/sfr_bro.xml").getroot()
    layer_element = xml.find(
        "dispatchDocument/SFR_O/soilFaceDescription/sfrcom:SoilFaceDescription/sfrcom:soilProfile/sfrcom:SoilProfile",
        xml.nsmap,
    )
    result = resolvers.process_sfr_data(layer_element, None)
    assert result == {
        "upperBoundary": [
            "0.000",
            "0.070",
            "0.250",
            "0.550",
            "0.830",
            "1.200",
            "1.400",
        ],
        "lowerBoundary": [
            "0.070",
            "0.250",
            "0.550",
            "0.830",
            "1.200",
            "1.400",
            "1.600",
        ],
        "soilNameNEN5104": [
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "matigSiltigeKlei",
            "sterkSiltigeKlei",
            "uiterstSiltigeKlei",
            "matigZandigeKlei",
            "sterkZandigeKlei",
        ],
    }

    result = resolvers.process_sfr_data(
        layer_element,
        [
            "upperBoundary",
            "lowerBoundary",
            "soilNameNEN5104",
            "colour",
            "estimatedClayContent",
            "nonexistingAttribute",  # Make sure non-existing attributes does not raise error
        ],
    )
    assert result == {
        "upperBoundary": [
            "0.000",
            "0.070",
            "0.250",
            "0.550",
            "0.830",
            "1.200",
            "1.400",
        ],
        "lowerBoundary": [
            "0.070",
            "0.250",
            "0.550",
            "0.830",
            "1.200",
            "1.400",
            "1.600",
        ],
        "soilNameNEN5104": [
            "sterkSiltigeKlei",
            "sterkSiltigeKlei",
            "matigSiltigeKlei",
            "sterkSiltigeKlei",
            "uiterstSiltigeKlei",
            "matigZandigeKlei",
            "sterkZandigeKlei",
        ],
        "colour": [
            "zwartBruin",
            "donkergrijs",
            "donkergrijs",
            "grijs",
            "grijs",
            None,
            None,
        ],
        "estimatedClayContent": ["28", "32", "48", "25", "20", "12", "8"],
        "nonexistingAttribute": [None, None, None, None, None, None, None],
    }
